from __future__ import annotations

from typing import Optional, Tuple, Literal

import torch
from scipy import stats as scipy_stats
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from einops import rearrange

from .in_and_out import PatchEmbed, ToPixel
from .pos_embed import (
    get_sincos_pos_embed,
    get_rope_tensor_2d,
    get_rope_tensor_3d,
    apply_rotary_emb,
)


PosType = Literal["none", "learned", "sincos"]
MaskType = Literal["none", "mae_random"]


class Mlp(nn.Module):
    def __init__(self, in_features: int, hidden_features: Optional[int] = None, out_features: Optional[int] = None,
                 act_layer=nn.GELU, drop: float = 0.0) -> None:
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x: Tensor) -> Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class DropPath(nn.Module):
    def __init__(self, drop_prob: float = 0.0) -> None:
        super().__init__()
        self.drop_prob = drop_prob

    def forward(self, x: Tensor) -> Tensor:
        if self.drop_prob == 0.0 or not self.training:
            return x
        keep_prob = 1 - self.drop_prob
        shape = (x.shape[0],) + (1,) * (x.ndim - 1)
        random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
        random_tensor.floor_()
        return x.div(keep_prob) * random_tensor


class Attention(nn.Module):
    """
    Multi-head attention with optional 2D/3D RoPE applied to a contiguous
    range of "image tokens" in the sequence.
    """

    def __init__(
        self,
        dim: int,
        num_heads: int = 8,
        qkv_bias: bool = True,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
    ) -> None:
        super().__init__()
        assert dim % num_heads == 0, f"dim % num_heads != 0, got {dim} and {num_heads}"
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(
        self,
        x: Tensor,
        rope_tensor: Optional[Tensor] = None,
        img_token_start: int = 0,
        img_token_end: Optional[int] = None,
    ) -> Tensor:
        bsz, n_ctx, ch = x.shape
        qkv = self.qkv(x)
        q, k, v = rearrange(qkv, "b n (qkv h d) -> qkv b h n d", qkv=3, h=self.num_heads).unbind(0)

        # Apply RoPE only to image tokens if provided
        if rope_tensor is not None:
            if img_token_end is None:
                img_token_end = n_ctx
            q_img = q[:, :, img_token_start:img_token_end]
            k_img = k[:, :, img_token_start:img_token_end]

            # rope_tensor shape: (num_img_tokens, head_dim * 2)
            num_img_tokens = img_token_end - img_token_start
            if rope_tensor.shape[0] < num_img_tokens:
                raise ValueError(
                    f"RoPE tensor has {rope_tensor.shape[0]} positions but need {num_img_tokens} "
                    f"for image tokens (seq_len={n_ctx}, img_token_start={img_token_start}, img_token_end={img_token_end})."
                )
            rope_slice = rope_tensor[:num_img_tokens]
            q_rot = apply_rotary_emb(q_img, rope_slice)
            k_rot = apply_rotary_emb(k_img, rope_slice)

            q = torch.cat([q[:, :, :img_token_start], q_rot, q[:, :, img_token_end:]], dim=2)
            k = torch.cat([k[:, :, :img_token_start], k_rot, k[:, :, img_token_end:]], dim=2)

        q = q * self.scale
        attn = q @ k.transpose(-2, -1)
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        x = attn @ v
        x = x.transpose(1, 2).reshape(bsz, n_ctx, ch)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class Block(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        drop: float = 0.0,
        attn_drop: float = 0.0,
        drop_path: float = 0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
    ) -> None:
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(
        self,
        x: Tensor,
        rope_tensor: Optional[Tensor] = None,
        img_token_start: int = 0,
        img_token_end: Optional[int] = None,
    ) -> Tensor:
        x = x + self.drop_path(self.attn(self.norm1(x), rope_tensor=rope_tensor,
                                         img_token_start=img_token_start, img_token_end=img_token_end))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class GenericViTEncoder(nn.Module):
    """
    General-purpose ViT-style encoder:

    - 2D or 3D PatchEmbed
    - optional CLS / prefix tokens
    - optional latent tokens
    - absolute pos (none / learned / sincos)
    - optional 2D/3D RoPE
    - optional MAE-style random masking

    Returns token sequence; caller can decide how to interpret it.
    """

    def __init__(
        self,
        img_size: int | Tuple[int, ...] = 224,
        patch_size: int | Tuple[int, ...] = 16,
        in_channels: int = 3,
        embed_dim: int = 768,
        depth: int = 12,
        num_heads: int = 12,
        mlp_ratio: float = 4.0,
        pos_type: PosType = "learned",
        use_rope: bool = True,
        dims: int = 2,
        num_prefix_tokens: int = 0,
        num_latent_tokens: int = 0,
        drop_path_rate: float = 0.0,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        qkv_bias: bool = True,
        masking: MaskType = "none",
        mask_ratio_mu: float = 0.0,
        mask_ratio_std: float = 0.15,
        mask_ratio_max: float = 0.75,
        mask_ratio_min: float = 0.05,
        double_z: bool = False,
    ) -> None:
        super().__init__()

        self.mask_ratio_mu = mask_ratio_mu
        self.mask_ratio_std = mask_ratio_std
        self.mask_ratio_max = mask_ratio_max
        self.mask_ratio_min = mask_ratio_min

        self.patch_embed = PatchEmbed(
            to_embed="conv",
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_channels,
            embed_dim=embed_dim,
        )
        self.embed_dim = embed_dim
        # For VQModel compatibility (used as z_channels)
        self.z_channels = embed_dim
        self.vae_stride = patch_size
        self.num_prefix_tokens = num_prefix_tokens
        self.num_latent_tokens = num_latent_tokens
        self.dims = self.patch_embed.dims
        self.num_img_tokens = self.patch_embed.num_patches

        # Prefix tokens (e.g. CLS)
        if self.num_prefix_tokens > 0:
            self.prefix_tokens = nn.Parameter(torch.zeros(1, self.num_prefix_tokens, self.embed_dim))
            nn.init.trunc_normal_(self.prefix_tokens, std=0.02)
        else:
            self.prefix_tokens = None

        # Latent tokens (e.g. MAETok latent tokens)
        if self.num_latent_tokens > 0:
            self.latent_tokens = nn.Parameter(torch.zeros(1, self.num_latent_tokens, self.embed_dim))
            nn.init.trunc_normal_(self.latent_tokens, std=0.02)
            self.latent_pos_embed = nn.Parameter(torch.zeros(1, self.num_latent_tokens, self.embed_dim))
            nn.init.trunc_normal_(self.latent_pos_embed, std=0.02)
        else:
            self.latent_tokens = None
            self.latent_pos_embed = None

        # Absolute positional embeddings
        if pos_type == "none":
            self.pos_embed = None
        elif pos_type == "learned":
            self.pos_embed = nn.Parameter(torch.zeros(1, self.num_prefix_tokens + self.num_img_tokens, self.embed_dim))
            nn.init.trunc_normal_(self.pos_embed, std=0.02)
        elif pos_type == "sincos":
            # build fixed sin-cos positions for image tokens and prepend zeros for prefix
            grid_size = self.patch_embed.grid_size
            if self.dims == 2:
                sincos = get_sincos_pos_embed(self.embed_dim, grid_size, dims=2)
            else:
                if self.embed_dim % 3 != 0:
                    raise ValueError(
                        f"3D sin-cos positional embedding requires embed_dim divisible by 3, got {self.embed_dim}"
                    )
                sincos = get_sincos_pos_embed(self.embed_dim, grid_size, dims=3)
            pos = torch.from_numpy(sincos).float()  # (num_img_tokens, dim)
            if self.num_prefix_tokens > 0:
                prefix_pos = torch.zeros(self.num_prefix_tokens, self.embed_dim)
                pos = torch.cat([prefix_pos, pos], dim=0)
            self.pos_embed = nn.Parameter(pos.unsqueeze(0), requires_grad=False)
        else:
            raise ValueError(f"Unknown pos_type: {pos_type}")

        # Masking config (MAE-style)
        self.masking = masking
        if self.masking == "mae_random":
            # Avoid division by zero when std is 0; use uniform [min, max] if std too small
            std = max(self.mask_ratio_std, 1e-8)
            a = (self.mask_ratio_min - self.mask_ratio_mu) / std
            b = (self.mask_ratio_max - self.mask_ratio_mu) / std
            self.mask_ratio_generator = scipy_stats.truncnorm(a, b, loc=self.mask_ratio_mu, scale=std)
            # simple mask token
            self.mask_token = nn.Parameter(torch.zeros(1, 1, self.embed_dim))
            nn.init.trunc_normal_(self.mask_token, std=0.02)
        else:
            self.mask_token = None

        # RoPE buffers (2D or 3D)
        self.use_rope = use_rope
        if self.use_rope:
            head_dim = self.embed_dim // num_heads
            grid = self.patch_embed.grid_size
            if self.dims == 2:
                rope = get_rope_tensor_2d(head_dim, grid[0], grid[1])
                self.grid_h, self.grid_w = grid
                self.register_buffer("rope_tensor", rope, persistent=False)
            else:
                rope = get_rope_tensor_3d(head_dim, grid[0], grid[1], grid[2])
                self.grid_d, self.grid_h, self.grid_w = grid
                self.register_buffer("rope_tensor", rope, persistent=False)
        else:
            self.register_buffer("rope_tensor", None, persistent=False)

        # Transformer blocks
        dpr = torch.linspace(0, drop_path_rate, depth).tolist()
        self.blocks = nn.ModuleList(
            [
                Block(
                    dim=self.embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    drop=proj_drop,
                    attn_drop=attn_drop,
                    drop_path=dpr[i],
                )
                for i in range(depth)
            ]
        )
        self.norm = nn.LayerNorm(self.embed_dim if not double_z else 2 * self.embed_dim)
        self.variational_z_doubler = nn.Linear(self.embed_dim, 2 * self.embed_dim) if double_z else nn.Identity()

    # --- Masking ---
    def _mae_random_mask(self, x: Tensor, mask_ratio: Optional[float] = None):
        """
        x: (B, L, C) full sequence including prefix + image + latent tokens.
        We only mask *image* tokens (between prefix and before latent).
        Returns only visible tokens for efficient processing.
        """
        bsz, seq_len, _ = x.shape
        img_start = self.num_prefix_tokens
        img_end = self.num_prefix_tokens + self.num_img_tokens
        img_tokens = x[:, img_start:img_end]

        if mask_ratio is None:
            mask_ratio = float(self.mask_ratio_max)

        if mask_ratio <= 0.0:
            mask = torch.zeros(bsz, self.num_img_tokens, device=x.device)
            return x, mask, None, None

        len_keep = int(self.num_img_tokens * (1.0 - mask_ratio))

        noise = torch.rand(bsz, self.num_img_tokens, device=x.device)
        ids_shuffle = torch.argsort(noise, dim=1)
        ids_restore = torch.argsort(ids_shuffle, dim=1)
        ids_keep = ids_shuffle[:, :len_keep]  # Original positions of visible tokens (in shuffled order)

        # Extract only visible tokens (not mask tokens)
        visible_img_tokens = torch.gather(img_tokens, 1, ids_keep[..., None].repeat(1, 1, img_tokens.shape[-1]))

        # Create mask for full sequence (for decoder reconstruction)
        mask = torch.ones(bsz, self.num_img_tokens, device=x.device)
        mask[:, :len_keep] = 0
        mask = torch.gather(mask, 1, ids_restore)

        # Build sequence with only visible tokens: prefix + visible_img + latent
        tokens_list = []
        if self.num_prefix_tokens > 0:
            tokens_list.append(x[:, :img_start])  # prefix tokens
        tokens_list.append(visible_img_tokens)  # only visible image tokens
        if self.num_latent_tokens > 0:
            tokens_list.append(x[:, img_end:])  # latent tokens
        
        x_visible = torch.cat(tokens_list, dim=1) if len(tokens_list) > 1 else tokens_list[0]
        
        return x_visible, mask, ids_restore, ids_keep

    def forward(self, x: Tensor, mask_ratio: Optional[float] = None):
        """
        x: (B, C, H, W) or (B, C, D, H, W)
        Returns:
          tokens: (B, L, C)
          optional dict with mask / ids_restore for MAE-style usage.
        """
        B = x.shape[0]
        x = self.patch_embed(x)  # (B, num_patches, embed_dim)

        tokens = []
        if self.prefix_tokens is not None:
            tokens.append(self.prefix_tokens.expand(B, -1, -1))
        tokens.append(x)
        seq = torch.cat(tokens, dim=1) if tokens else x  # prefix + image

        # absolute pos for prefix+image tokens
        if self.pos_embed is not None:
            seq = seq + self.pos_embed

        # append latent tokens (with their pos)
        if self.latent_tokens is not None:
            latent = self.latent_tokens.expand(B, -1, -1)
            if self.latent_pos_embed is not None:
                latent = latent + self.latent_pos_embed
            seq = torch.cat([seq, latent], dim=1)

        # masking (on image tokens only)
        mask = None
        ids_restore = None
        ids_keep = None
        num_visible_tokens = None
        if self.masking == "mae_random":
            mask_ratio = self.mask_ratio_generator.rvs(1)[0] if mask_ratio is None else mask_ratio
            print(f"mask_ratio: {mask_ratio}")
            seq, mask, ids_restore, ids_keep = self._mae_random_mask(seq, mask_ratio=mask_ratio)
            # After masking, seq contains only visible tokens
            # Calculate number of visible image tokens
            num_visible_tokens = ids_keep.shape[1] if ids_keep is not None else int(self.num_img_tokens * (1.0 - (mask_ratio if mask_ratio is not None else self.mask_ratio_max)))

        # RoPE tensor (for image tokens only)
        rope_tensor = None
        img_token_start = self.num_prefix_tokens
        # When masking, img_token_end is adjusted to only visible tokens
        if num_visible_tokens is not None:
            img_token_end = self.num_prefix_tokens + num_visible_tokens
        else:
            img_token_end = self.num_prefix_tokens + self.num_img_tokens

        if self.use_rope and self.rope_tensor is not None:
            rope_tensor_full = self.rope_tensor.to(seq.device, dtype=seq.dtype)
            if num_visible_tokens is not None and ids_keep is not None:
                # Extract RoPE for visible tokens: ids_keep contains original positions of visible tokens
                # Use first batch's ids_keep to extract RoPE (same for all batches)
                rope_tensor = rope_tensor_full[ids_keep[0]]
            else:
                rope_tensor = rope_tensor_full

        for blk in self.blocks:
            seq = blk(seq, rope_tensor=rope_tensor, img_token_start=img_token_start, img_token_end=img_token_end)

        seq = self.variational_z_doubler(seq)
            
        seq = self.norm(seq)

        # Extract tokens for output: exclude prefix tokens (CLS) from latent space
        # When masking, return only visible tokens; otherwise return all image tokens
        img_start = self.num_prefix_tokens
        
        if self.num_latent_tokens > 0:
            # Return only latent tokens (B, num_latent_tokens, C)
            tokens_out = seq[:, -self.num_latent_tokens:]
        else:
            # Return only visible image tokens (when masking) or all image tokens (when not masking)
            # Exclude prefix tokens
            tokens_out = seq[:, img_start:img_token_end]

        aux = {"mask": mask, "ids_restore": ids_restore}
        return tokens_out, aux


class GenericViTDecoder(nn.Module):
    """
    General-purpose ViT-style decoder:

    - accepts latent tokens (optionally plus mask restoration indices)
    - can insert mask tokens and restore full image token sequence (MAE-style)
    - absolute pos + optional 2D/3D RoPE
    - generic ToPixel head for reconstruction
    """

    def __init__(
        self,
        img_size: int | Tuple[int, ...] = 224,
        patch_size: int | Tuple[int, ...] = 16,
        out_channels: int = 3,
        embed_dim: int = 768,
        depth: int = 12,
        num_heads: int = 12,
        mlp_ratio: float = 4.0,
        pos_type: PosType = "learned",
        use_rope: bool = True,
        dims: int = 2,
        num_prefix_tokens: int = 0,
        num_latent_tokens: int = 0,
        drop_path_rate: float = 0.0,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        qkv_bias: bool = True,
        to_pixel: str = "conv",
        token_dim: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.embed_dim = embed_dim

        # number of image tokens from img_size / patch_size
        self.to_pixel = ToPixel(
            to_pixel=to_pixel,
            img_size=img_size,
            out_channels=out_channels,
            in_dim=embed_dim,
            patch_size=patch_size,
        )
        self.num_img_tokens = self.to_pixel.num_patches

        self.num_prefix_tokens = num_prefix_tokens
        self.num_latent_tokens = num_latent_tokens

        # Prefix tokens (independent from encoder)
        if self.num_prefix_tokens > 0:
            self.prefix_tokens = nn.Parameter(torch.zeros(1, self.num_prefix_tokens, self.embed_dim))
            nn.init.trunc_normal_(self.prefix_tokens, std=0.02)
        else:
            self.prefix_tokens = None

        # mask token for MAE-style reconstruction
        self.mask_token = nn.Parameter(torch.zeros(1, 1, self.embed_dim))
        nn.init.trunc_normal_(self.mask_token, std=0.02)

        # Absolute positional embeddings
        if pos_type == "none":
            self.pos_embed = None
        elif pos_type == "learned":
            self.pos_embed = nn.Parameter(torch.zeros(1, self.num_prefix_tokens + self.num_img_tokens, self.embed_dim))
            nn.init.trunc_normal_(self.pos_embed, std=0.02)
        elif pos_type == "sincos":
            grid_size = self.to_pixel.grid_size
            if self.to_pixel.dims == 2:
                sincos = get_sincos_pos_embed(self.embed_dim, grid_size, dims=2)
            else:
                sincos = get_sincos_pos_embed(self.embed_dim, grid_size, dims=3)
            pos = torch.from_numpy(sincos).float()
            if self.num_prefix_tokens > 0:
                prefix_pos = torch.zeros(self.num_prefix_tokens, self.embed_dim)
                pos = torch.cat([prefix_pos, pos], dim=0)
            self.pos_embed = nn.Parameter(pos.unsqueeze(0), requires_grad=False)
        else:
            raise ValueError(f"Unknown pos_type: {pos_type}")

        # RoPE buffers
        self.use_rope = use_rope
        if self.use_rope:
            head_dim = self.embed_dim // num_heads
            grid = self.to_pixel.grid_size
            if self.to_pixel.dims == 2:
                rope = get_rope_tensor_2d(head_dim, grid[0], grid[1])
                self.grid_h, self.grid_w = grid
                self.register_buffer("rope_tensor", rope, persistent=False)
            else:
                rope = get_rope_tensor_3d(head_dim, grid[0], grid[1], grid[2])
                self.grid_d, self.grid_h, self.grid_w = grid
                self.register_buffer("rope_tensor", rope, persistent=False)
        else:
            self.register_buffer("rope_tensor", None, persistent=False)

        # Input projection: tokens -> embed_dim
        in_dim = token_dim or self.embed_dim
        if in_dim != self.embed_dim:
            self.decoder_embed = nn.Linear(in_dim, self.embed_dim)
        else:
            self.decoder_embed = nn.Identity()

        # Transformer blocks
        dpr = torch.linspace(0, drop_path_rate, depth).tolist()
        self.blocks = nn.ModuleList(
            [
                Block(
                    dim=self.embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    drop=proj_drop,
                    attn_drop=attn_drop,
                    drop_path=dpr[i],
                )
                for i in range(depth)
            ]
        )
        self.norm = nn.LayerNorm(self.embed_dim)

    def get_last_layer(self):
        """Return the last layer weights for adaptive GAN loss weighting."""
        return self.to_pixel.get_last_layer()

    def forward(self, tokens: Tensor, ids_restore: Optional[Tensor] = None) -> Tensor:
        """
        tokens: (B, L, C_token) latent tokens or visible tokens.
        ids_restore: if provided, MAE-style unshuffling and mask-token insertion.
        """
        x = self.decoder_embed(tokens)
        B, L, _ = x.shape

        # If ids_restore provided, create masked sequence of image tokens
        if ids_restore is not None:
            num_img_tokens = ids_restore.shape[1]
            num_mask_tokens = num_img_tokens - L
            if num_mask_tokens < 0:
                raise ValueError(f"ids_restore expects at least {L} tokens, got {num_img_tokens}")
            mask_tokens = self.mask_token.expand(B, num_mask_tokens, -1)
            x_ = torch.cat([x, mask_tokens], dim=1)  # (B, L + num_mask_tokens, C)
            x = torch.gather(x_, 1, ids_restore.unsqueeze(-1).expand(-1, -1, x_.shape[-1]))

        # Prepend decoder's own prefix tokens (independent from encoder)
        seq = x
        if self.prefix_tokens is not None:
            prefix = self.prefix_tokens.expand(B, -1, -1)
            seq = torch.cat([prefix, seq], dim=1)  # (B, num_prefix_tokens + L, C)

        # Add absolute pos on prefix+image tokens
        if self.pos_embed is not None:
            seq = seq + self.pos_embed

        img_token_start = self.num_prefix_tokens
        img_token_end = self.num_prefix_tokens + self.num_img_tokens

        rope_tensor = None
        if self.use_rope and self.rope_tensor is not None:
            rope_tensor = self.rope_tensor.to(seq.device, dtype=seq.dtype)

        for blk in self.blocks:
            seq = blk(seq, rope_tensor=rope_tensor, img_token_start=img_token_start, img_token_end=img_token_end)

        seq = self.norm(seq)

        # Drop prefix tokens before ToPixel
        seq_img = seq[:, img_token_start:img_token_end]
        out = self.to_pixel(seq_img)
        return out




# Adopted from LDM's KL-VAE: https://github.com/CompVis/latent-diffusion
import torch
from torch import nn
from medlat.utils import init_from_ckpt
from typing import Optional, Sequence, Union, List, Any, Dict, Tuple
from medlat.first_stage.discrete.modules.ldm_modules import get_conv_layer
from medlat.modules.alignments import AlignmentModule
from einops import rearrange
__all__ = ["VQModel", "VQModelTransformer"]


class VQModel(nn.Module):
    def __init__(
        self,
        encoder: nn.Module,
        decoder: nn.Module,
        quantizer: nn.Module,
        alignment: AlignmentModule = None,
        ckpt_path=None,
        # Additional parameters
        quant_conv_ks=1,   ## in var its 3, but VQVAE / VQGAN uses 1
        pre_post_layer="conv",
    ):
        super().__init__()
        self.embed_dim = quantizer.e_dim
        self.n_embed = quantizer.n_e
        self.dims = getattr(encoder, "dims", 2)
        self.encoder = encoder
        self.decoder = decoder
        self.alignment = alignment
        self.z_channels = getattr(encoder, "z_channels", None)
        self.quantizer = quantizer

        if self.z_channels is None:
            raise ValueError(f"Encoder {encoder.__class__.__name__} must define z_channels.")

        self.vae_stride = getattr(encoder, "vae_stride", None)
        if pre_post_layer == "conv":
            conv_layer = get_conv_layer(self.dims)
            self.quant_conv = conv_layer(self.z_channels, self.embed_dim, quant_conv_ks, stride=1, padding=quant_conv_ks//2)
            self.post_quant_conv = conv_layer(self.embed_dim, self.z_channels, quant_conv_ks, stride=1, padding=quant_conv_ks//2)
        elif pre_post_layer == "none":
            self.quant_conv = nn.Identity()
            self.post_quant_conv = nn.Identity()
        else:
            raise ValueError(f"Invalid pre_post_layer: {pre_post_layer}")


        if ckpt_path is not None: 
            init_from_ckpt(self, ckpt_path)

    # def _check_msrq_features(self, method_name):
    #     if not self.quantizer.has_msrq_features:
    #         raise NotImplementedError(
    #             f"Method {method_name} requires MSRQ features. Please initialize VQModel_Combined with "
    #             "v_patch_nums, quant_resi, share_quant_resi, and using_znorm parameters."
    #         )

    def lock_parameters(self):
        """Lock the parameters of the model to prevent them from being updated during training."""
        for param in self.parameters():
            param.requires_grad = False

    def encode(self, x):
        h = self.encoder(x)
        # Allow encoders that return (features, aux)
        h = self.quant_conv(h)
        quant, emb_loss, info = self.quantizer(h)
        return quant, emb_loss, info

    def quantize(self, h):
        quant, emb_loss, info = self.quantizer(h)
        return quant, emb_loss, info
    
    def encode_to_prequant(self, x):
        h = self.encoder(x)
        h = self.quant_conv(h)
        return h, None, None

    def decode_from_prequant(self, h):
        quant, emb_loss, info = self.quantizer(h)
        return self.decode(quant)
        
    def decode(self, quant):
        quant = self.post_quant_conv(quant)
        dec = self.decoder(quant)
        return dec

    def decode_code(self, code_b, out_shape=None):
        quant_b = self.quantizer.get_codebook_entry(code_b, shape=out_shape)
        # Move channel dimension (which is last) to the second
        if quant_b.dim() == 4:
            # (B, H, W, C) -> (B, C, H, W)
            quant_b = quant_b.permute(0, 3, 1, 2).contiguous()
        elif quant_b.dim() == 5:
            # (B, D, H, W, C) -> (B, C, D, H, W)
            quant_b = quant_b.permute(0, 4, 1, 2, 3).contiguous()
        dec = self.decode(quant_b)
        return dec

    def forward(self, input, return_pred_indices=False):
        quant, diff, (_,_,ind) = self.encode(input)
        dec = self.decode(quant)

        if self.alignment is not None:
            alignment_loss, _ = self.alignment(quant, input)
            diff = diff + alignment_loss
            
        if return_pred_indices:
            return dec, diff, ind
        return dec, diff

    # # VAR/MSQR-specific methods
    # def fhat_to_img(self, f_hat: torch.Tensor):
    #     """Convert quantized features to image"""
    #     self._check_msrq_features('fhat_to_img')
    #     return self.decoder(self.post_quant_conv(f_hat)).clamp_(-1, 1)
    
    # def img_to_idxBl(self, inp_img_no_grad: torch.Tensor, v_patch_nums: Optional[Sequence[Union[int, Tuple[int, int]]]] = None) -> List[torch.LongTensor]:
    #     """Convert image to multi-scale indices"""
    #     self._check_msrq_features('img_to_idxBl')
    #     h = self.encoder(inp_img_no_grad)
    #     if isinstance(h, tuple):
    #         h = h[0]
    #     f = self.quant_conv(h)
    #     return self.quantizer.f_to_idxBl_or_fhat(f, to_fhat=False, v_patch_nums=v_patch_nums)
    
    # def idxBl_to_img(self, ms_idx_Bl: List[torch.Tensor], same_shape: bool, last_one=False) -> Union[List[torch.Tensor], torch.Tensor]:
    #     """Convert multi-scale indices to image"""
    #     self._check_msrq_features('idxBl_to_img')
    #     B = ms_idx_Bl[0].shape[0]
    #     ms_h_BChw = []
    #     for idx_Bl in ms_idx_Bl:
    #         l = idx_Bl.shape[1]
    #         pn = round(l ** 0.5)
    #         ms_h_BChw.append(self.quantizer.embedding(idx_Bl).transpose(1, 2).view(B, self.embed_dim, pn, pn))
    #     return self.embed_to_img(ms_h_BChw=ms_h_BChw, all_to_max_scale=same_shape, last_one=last_one)
    
    # def embed_to_img(self, ms_h_BChw: List[torch.Tensor], all_to_max_scale: bool, last_one=False) -> Union[List[torch.Tensor], torch.Tensor]:
    #     """Convert embeddings to image"""
    #     self._check_msrq_features('embed_to_img')
    #     if last_one:
    #         return self.decoder(self.post_quant_conv(self.quantizer.embed_to_fhat(ms_h_BChw, all_to_max_scale=all_to_max_scale, last_one=True))).clamp_(-1, 1)
    #     else:
    #         return [self.decoder(self.post_quant_conv(f_hat)).clamp_(-1, 1) for f_hat in self.quantizer.embed_to_fhat(ms_h_BChw, all_to_max_scale=all_to_max_scale, last_one=False)]
    
    # def img_to_reconstructed_img(self, x, v_patch_nums: Optional[Sequence[Union[int, Tuple[int, int]]]] = None, last_one=False) -> List[torch.Tensor]:
    #     """Convert image to reconstructed image at multiple scales"""
    #     self._check_msrq_features('img_to_reconstructed_img')
    #     h = self.encoder(x)
    #     if isinstance(h, tuple):
    #         h = h[0]
    #     f = self.quant_conv(h)
    #     ls_f_hat_BChw = self.quantizer.f_to_idxBl_or_fhat(f, to_fhat=True, v_patch_nums=v_patch_nums)
    #     if last_one:
    #         return self.decoder(self.post_quant_conv(ls_f_hat_BChw[-1])).clamp_(-1, 1)
    #     else:
    #         return [self.decoder(self.post_quant_conv(f_hat)).clamp_(-1, 1) for f_hat in ls_f_hat_BChw]

    # def load_state_dict(self, state_dict: Dict[str, Any], strict=True, assign=False):
    #     if 'quantizer.ema_vocab_hit_SV' in state_dict and state_dict['quantizer.ema_vocab_hit_SV'].shape[0] != self.quantizer.ema_vocab_hit_SV.shape[0]:
    #         state_dict['quantizer.ema_vocab_hit_SV'] = self.quantizer.ema_vocab_hit_SV
    #     return super().load_state_dict(state_dict=state_dict, strict=strict, assign=assign)


class VQModelTransformer(nn.Module):
    def __init__(
        self,
        encoder: nn.Module,
        decoder: nn.Module,
        quantizer: nn.Module,
        alignment: AlignmentModule = None,
        ckpt_path=None,
        pre_post_layer="linear",
    ):
        super().__init__()
        self.embed_dim = quantizer.e_dim
        self.n_embed = quantizer.n_e
        self.dims = getattr(encoder, "dims", 2)
        self.encoder = encoder
        self.decoder = decoder
        self.alignment = alignment
        self.z_channels = getattr(encoder, "z_channels", None)
        self.quantizer = quantizer

        if self.z_channels is None:
            raise ValueError(f"Encoder {encoder.__class__.__name__} must define z_channels.")

        self.vae_stride = getattr(encoder, "vae_stride", None)
        if pre_post_layer == "linear":
            self.quant_conv = nn.Linear(self.z_channels, self.embed_dim)
            self.post_quant_conv = nn.Linear(self.embed_dim, self.z_channels)
        elif pre_post_layer == "none":
            self.quant_conv = nn.Identity()
            self.post_quant_conv = nn.Identity()
        else:
            raise ValueError(f"Invalid pre_post_layer: {pre_post_layer}")


        if ckpt_path is not None: 
            init_from_ckpt(self, ckpt_path)


    def lock_parameters(self):
        """Lock the parameters of the model to prevent them from being updated during training."""
        for param in self.parameters():
            param.requires_grad = False

    def encode(self, x):
        h, aux = self.encoder(x)
        h = self.quant_conv(h)
        quant, emb_loss, info = self.quantizer(h)
        return quant, emb_loss, info, aux

    def quantize(self, h):
        quant, emb_loss, info = self.quantizer(h)
        return quant, emb_loss, info

    def encode_to_prequant(self, x):
        h, aux = self.encoder(x)
        h = self.quant_conv(h)
        return h, None, None, aux

    def decode_from_prequant(self, h, aux=None):
        quant, emb_loss, info = self.quantizer(h)
        return self.decode(quant, aux=aux)
        
    def decode(self, quant, aux=None):
        quant = self.post_quant_conv(quant)
        dec = self.decoder(quant, ids_restore=aux["ids_restore"] if aux is not None else None)
        return dec

    def decode_code(self, code_b, out_shape=None, aux=None):
        quant_b = self.quantizer.get_codebook_entry(code_b, shape=out_shape)
        # Move channel dimension (which is last) to the second
        if quant_b.dim() == 4:
            # (B, H, W, C) -> (B, C, H, W)
            quant_b = quant_b.permute(0, 3, 1, 2).contiguous()
        elif quant_b.dim() == 5:
            # (B, D, H, W, C) -> (B, C, D, H, W)
            quant_b = quant_b.permute(0, 4, 1, 2, 3).contiguous()
        dec = self.decode(quant_b, aux=aux)
        return dec

    def forward(self, input, return_pred_indices=False):
        quant, diff, (_,_,ind), aux = self.encode(input)
        dec = self.decode(quant, aux=aux)

        if self.alignment is not None:
            alignment_loss, _ = self.alignment(quant, input)
            diff = diff + alignment_loss
            
        if return_pred_indices:
            return dec, diff, ind
        return dec, diff


import torch
from torch import nn
from typing import List, Sequence, Tuple, Union
from torch.utils.checkpoint import checkpoint
from timm.models.vision_transformer import Block
from typing import Optional

# from networks.unetr_blocks import UnetOutBlock, UnetrBasicBlock, UnetrPrUpBlock, UnetrUpBlock


import numpy as np

from .utils.imaging_model_related import Masker, sincos_pos_embed, patchify, unpatchify


class PatchEmbed(nn.Module):
    def __init__(self, im_shape: list[int], 
                 in_channels: int = 1, 
                 patch_size: list[int] = [1, 16, 16], 
                 out_channels: int = 256, 
                 flatten: bool = True, 
                 bias: bool = True,
                 norm_layer: Optional[nn.Module] = None, **kwargs):
        super().__init__()
        
        assert len(patch_size) == 3, "Patch size should be 3D"
        assert in_channels == 1, "Only supporting input channel size as 1"
        print(f"im_shape: {im_shape}, patch_size: {patch_size}")
        self.im_shape = im_shape
        if len(im_shape) == 3:
            self.im_shape = im_shape.unsqueeze(1) # (S, H, W) -> (S, 1, H, W)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.flatten = flatten
        
        self.proj = nn.Conv3d(in_channels, out_channels, kernel_size=patch_size, stride=patch_size, bias=bias)
        self.norm = norm_layer(out_channels) if norm_layer else nn.Identity()
        self.grid_size = (im_shape[0], 
                          im_shape[1] // patch_size[0], 
                          im_shape[2] // patch_size[1], 
                          im_shape[3] // patch_size[2]) # (S, t, h, w)
        self.num_patches = np.prod(self.grid_size)
        print(f"grid_size: {self.grid_size}, num_patches: {self.num_patches}")
    
    def forward(self, x):
        """ 
        input: (B, S, T, H, W)
        output: (B * S, out_channels, t, h, w) or (B, num_patches, out_channels) if flatten is True, 
                where num_patches = S * T * H * W / np.prod(patch_size)
        """
        x_ = x.reshape(-1, *self.im_shape[-3:]) # (B*S, T, H, W)
        x_ = x_.unsqueeze(1) # (B*S, 1, T, H, W)
        x_ = self.proj(x_) # (B*S, out_channels, t, h, w)
        
        if self.flatten:
            x__ = x_.flatten(2) # (B*S, out_channels, t*h*w)
            x__ = x__.moveaxis(1, -1) # (B*S, t*h*w, out_channels)
            x_ = x__.reshape(x.shape[0], -1, x__.shape[-1]) # (B, S*t*h*w, out_channels)
        else:
            x_ = x_.moveaxis(1, -1)
            
        x = self.norm(x_)
        return x

class ImagingMaskedEncoder(nn.Module):
    def __init__(
        self,
        patch_embed_cls: str = "PatchEmbed",
        patch_size: Tuple[int, ...] = (5, 8, 8),
        patch_in_channels: int = 1,
        pixel_unshuffle_scale: int = 1,
        mask_type: str = "random",
        mask_ratio: float = 0.0,
        circular_pe: bool = False,
        use_enc_pe: bool = True,
        mask_loss: bool = True,
        shift_size: Tuple[int, ...] = (0, 0, 0),
        enc_embed_dim: int = 1025, # should be divisible by 8 or 6 for one modality, +1 for two modalities
        enc_depth: int = 6,
        enc_num_heads: int = 5,
        mlp_ratio: float = 4.0,
        grad_checkpointing: bool = False,
        img_shape: tuple = None,
        use_both_axes: bool = False,
        *args, **kwargs
    ):
        super().__init__()
        self.enc_embed_dim = enc_embed_dim
        self.patch_size = patch_size
        self.circular_pe = circular_pe
        self.use_enc_pe = use_enc_pe
        self.mask_loss = mask_loss
        self.shift_size = shift_size
        self.patch_embed_cls = globals()[patch_embed_cls]
        self.img_shape = img_shape
        self.patch_in_channels = patch_in_channels
        self.use_both_axes = use_both_axes
        self.patch_p_num = np.prod(patch_size) * self.patch_in_channels
        self.grad_checkpointing = grad_checkpointing
        self.patch_embed = PatchEmbed(img_shape, 
                                    in_channels=patch_in_channels, 
                                    patch_size=patch_size, 
                                    out_channels=enc_embed_dim, )
        print(f"patch_embed.num_patches: {self.patch_embed.num_patches}")
        self.cls_token = nn.Parameter(torch.zeros(1, 1, self.patch_embed.out_channels))
        self.enc_pos_embed = nn.Parameter(torch.zeros(1, 1 + self.patch_embed.num_patches, self.patch_embed.out_channels), 
                                      requires_grad=False)
        self.encoder = nn.ModuleList([Block(dim=self.patch_embed.out_channels, 
                                            num_heads=enc_num_heads, 
                                            mlp_ratio=mlp_ratio, 
                                            qkv_bias=True,)
                                      for i in range(enc_depth)])
        self.encoder_norm = nn.LayerNorm(self.patch_embed.out_channels)
        # --------------------------------------------------------------------------
        # MAE Masker
        self.using_masker = False if mask_ratio == 0.0 else True
        if self.using_masker:
            self.masker = Masker(mask_type=mask_type, 
                                 mask_ratio=mask_ratio, 
                                 grid_size=patch_embed.grid_size)
        self.initialize_parameters()
    
    def initialize_parameters(self):        
        # Initialize (and freeze) pos_embed by sin-cos embedding
        enc_pos_embed = sincos_pos_embed(self.enc_embed_dim, self.patch_embed.grid_size, cls_token=True,
                                        use_both_axes=self.use_both_axes, circular_pe=self.circular_pe)
        if not self.use_enc_pe:
            enc_pos_embed[:, 0] = 0
        self.enc_pos_embed.data.copy_(enc_pos_embed.unsqueeze(0))

        # Initialize patch_embed like nn.Linear (instead of nn.Conv2d)
        w = self.patch_embed.proj.weight.data
        torch.nn.init.xavier_uniform_(w.view([w.shape[0], -1]))
        # timm"s trunc_normal_(std=.02) is effectively normal_(std=0.02) as cutoff is too big (2.)
        if hasattr(self, "cls_token"):
            torch.nn.init.normal_(self.cls_token, std=.02)

        # Initialize nn.Linear and nn.LayerNorm
        self.apply(self._init_weights)
        
    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
            
    def forward(self, x):
        """Forward pass of encoder
        input: [B, S, T, H, W] torch.Tensor
        output:
            latent: [B, 1 + length * mask_ratio, embed_dim] torch.Tensor
            mask: [B, 1 + length * mask_ratio] torch.Tensor
            ids_restore: [B, 1 + length * mask_ratio] torch.Tensor
        """
        # Embed patches: (B, S, T, H, W) -> (B, S * T * num_patches, embed_dim)
        print(f"x.shape: {x.shape}")
        x = self.patch_embed(x)
        print(f"x.shape: {x.shape}")
        
        # Add positional embedding: (B, S * T * num_patches, embed_dim)
        if self.use_enc_pe:
            enc_pos_embed = self.enc_pos_embed.repeat(x.shape[0], 1, 1)
            x = x + enc_pos_embed[:, 1:, :]
            cls_token = self.cls_token + enc_pos_embed[:, :1, :] # (1, 1, embed_dim)
        else:
            cls_token = self.cls_token
        
        # Mask patches: length -> length * mask_ratio
        if self.using_masker:
            x, mask, ids_restore = self.masker(x)
        else:
            mask, ids_restore = None, None
            
        # Append cls token: (B, 1 + length * mask_ratio, em bed_dim)
        cls_tokens = cls_token.expand(x.shape[0], -1, -1) # (B, 1, embed_dim)
        x = torch.cat((cls_tokens, x), dim=1)
        
        # Apply transformer encoder
        if self.grad_checkpointing and not torch.jit.is_scripting():
            for blk in self.encoder:
                x = checkpoint(blk, x, use_reentrant=False)
        else:
            for blk in self.encoder:
                x = blk(x)
        x = self.encoder_norm(x)
        return x, mask, ids_restore

    def forward_with_skip_connection(self, x):
        # Embed patches: (B, S, T, H, W) -> (B, S * num_patches, embed_dim)
        x = self.patch_embed(x)

        # Add positional embedding: (B, S * num_patches, embed_dim)
        if self.use_enc_pe:
            enc_pos_embed = self.enc_pos_embed.repeat(x.shape[0], 1, 1)
            x = x + enc_pos_embed[:, 1:, :]
            cls_token = self.cls_token + enc_pos_embed[:, :1, :] # (1, 1, embed_dim)
        else:
            cls_token = self.cls_token
        
        # Append cls token: (B, 1 + length * mask_ratio, embed_dim)
        cls_tokens = cls_token.expand(x.shape[0], -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)
        
        # Apply MAE encoder
        hidden_latents = []
        if self.grad_checkpointing and not torch.jit.is_scripting():
            for blk in self.encoder:
                hidden_latents.append(x[:, 1:, :])
                x = checkpoint(blk, x, use_reentrant=False)
        else:
            for blk in self.encoder:
                hidden_latents.append(x[:, 1:, :])
                x = blk(x)
        encoder_output = self.encoder_norm(x)
        return encoder_output, hidden_latents


class Layer(nn.Module):
    def __init__(self, in_size, out_size, dropout=0.0, **kwargs):
        super(Layer, self).__init__()
        self.dropout = None
        if dropout > 0.0:
            self.dropout = nn.Dropout(dropout)
        self.in_size = in_size
        self.out_size = out_size

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError


class Relu(Layer):
    """A linear layer followed by a ReLU activation function."""
    def __init__(self, in_size, out_size, dropout=False, **kwargs):
        super(Relu, self).__init__(in_size, out_size, **kwargs)
        self.linear = nn.Linear(in_size, out_size)
        if dropout:
            self.dropout = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear(x)
        x = torch.relu(x)
        if self.dropout is not None:
            x = self.dropout(x)
        return x


class ViTDecoder(nn.Module):
    def __init__(self, dim, num_heads, depth, mlp_ratio, norm_layer=nn.LayerNorm, grad_checkpointing: bool = False):
        super(ViTDecoder, self).__init__()
        self.network = nn.ModuleList([
            Block(dim, num_heads, mlp_ratio, qkv_bias=True, norm_layer=norm_layer)
            for i in range(depth)])
        self.norm = norm_layer(dim)
        self.grad_checkpointing = grad_checkpointing
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Apply transformer decoder
        if self.grad_checkpointing and not torch.jit.is_scripting():
            for blk in self.network:
                x = checkpoint(blk, x, use_reentrant=False)
        else:
            for blk in self.network:
                x = blk(x)
        x = self.norm(x)
        return x
        

class LinearDecoder(nn.Module):
    def __init__(self, in_size, dim, depth, layer_type:Layer=Relu):
        super(LinearDecoder, self).__init__()
        self.network = nn.ModuleList([layer_type(in_size, dim // 2, dropout=0.1)])
        self.network.extend([
            layer_type(dim // 2 ** i, dim // 2 ** (i + 1), dropout=0.1) for i in range(1, depth)])
        self.fc = nn.Linear(dim // 2 ** depth, 1, bias=True)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if len(x.shape) == 3:
            x = x.flatten(start_dim=1)
        for layer in self.network:
            x = layer(x)
        x = self.fc(x)
        return x


class UNETR_decoder(nn.Module):
    """
    UNETR based on: "Hatamizadeh et al.,
    UNETR: Transformers for 3D Medical Image Segmentation <https://arxiv.org/abs/2103.10504>"
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        img_size: Union[Sequence[int], int],
        patch_size: Union[Sequence[int], int],
        upsample_kernel_sizes: Union[list, Sequence[int]],
        feature_size: int = 16,
        hidden_size: int = 768,
        norm_name: Union[Tuple, str] = "instance",
        conv_block: bool = True,
        res_block: bool = True,
        spatial_dims: int = 3,
    ) -> None:
        """
        Args:
            in_channels: dimension of input channels.
            out_channels: dimension of output channels.
            img_size: dimension of input image.
            feature_size: dimension of network feature size.
            hidden_size: dimension of hidden layer.
            norm_name: feature normalization type and arguments.
            conv_block: bool argument to determine if convolutional block is used.
            res_block: bool argument to determine if residual block is used.
            spatial_dims: number of spatial dims.

        Examples::

            # for single channel input 4-channel output with image size of (96,96,96), feature size of 32 and batch norm
            >>> net = UNETR(in_channels=1, out_channels=4, img_size=(96,96,96), feature_size=32, norm_name="batch")

             # for single channel input 4-channel output with image size of (96,96), feature size of 32 and batch norm
            >>> net = UNETR(in_channels=1, out_channels=4, img_size=96, feature_size=32, norm_name="batch", spatial_dims=2)

        """

        super().__init__()

        patch_size = (1, *patch_size)
        self.grid_size = tuple(img_d // p_d for img_d, p_d in zip(img_size, patch_size))
        self.hidden_size = hidden_size
        self.slice_num = img_size[0]
        self.output_channel = out_channels * self.slice_num # times slice num
        self.upsample_kernel_sizes = upsample_kernel_sizes
        assert len(self.upsample_kernel_sizes) == 3, "Only support UNETR decoder depth equals 3"
            
        self.encoder1 = UnetrBasicBlock(
            spatial_dims=spatial_dims,
            in_channels=in_channels,
            out_channels=feature_size,
            kernel_size=3,
            stride=1,
            norm_name=norm_name,
            res_block=res_block,
        )
        self.encoder2 = UnetrPrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=hidden_size,
            out_channels=feature_size * 2,
            num_layer=1,
            kernel_size=3,
            stride=1,
            upsample_kernel_size=self.upsample_kernel_sizes[1:],
            norm_name=norm_name,
            conv_block=conv_block,
            res_block=res_block,
        )
        self.encoder3 = UnetrPrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=hidden_size,
            out_channels=feature_size * 4,
            num_layer=0,
            kernel_size=3,
            stride=1,
            upsample_kernel_size=self.upsample_kernel_sizes[2],
            norm_name=norm_name,
            conv_block=conv_block,
            res_block=res_block,
        )
        self.decoder4 = UnetrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=hidden_size,
            out_channels=feature_size * 4,
            kernel_size=3,
            upsample_kernel_size=self.upsample_kernel_sizes[2],
            norm_name=norm_name,
            res_block=res_block,
        )
        self.decoder3 = UnetrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=feature_size * 4,
            out_channels=feature_size * 2,
            kernel_size=3,
            upsample_kernel_size=self.upsample_kernel_sizes[1],
            norm_name=norm_name,
            res_block=res_block,
        )
        self.decoder2 = UnetrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=feature_size * 2,
            out_channels=feature_size,
            kernel_size=3,
            upsample_kernel_size=self.upsample_kernel_sizes[0],
            norm_name=norm_name,
            res_block=res_block,
        )
        self.out = UnetOutBlock(spatial_dims=spatial_dims, in_channels=feature_size, out_channels=self.output_channel)

    def proj_feat(self, x, hidden_size, grid_size):
        new_view = (x.size(0), *grid_size, hidden_size)
        x = x.view(new_view)
        new_axes = (0, len(x.shape) - 1) + tuple(d + 1 for d in range(len(grid_size)))
        x = x.permute(new_axes).contiguous()
        return x

    def forward(self, x_in: torch.Tensor, x: torch.Tensor, hidden_states_out: List[torch.Tensor]) -> torch.Tensor:
        """Forward pass of UNETR decoder.

        Args:
            x_in (torch.Tensor): images in the shape of (batch, slice, time, height, width)
            x (torch.Tensor): latent features extracted from the encoder
            hidden_states_out (List[torch.Tensor]): the output of each layer of encoder

        Returns:
            torch.Tensor: segmentation probability in the same shape as x_in
        """
        enc1 = self.encoder1(x_in)
        x2 = hidden_states_out[0]
        proj_x2 = self.proj_feat(x2, self.hidden_size//self.grid_size[0], self.grid_size) # (batch, hidden_size, s, 2, 16, 16)
        proj_x2 = proj_x2.view(proj_x2.shape[0], -1, *self.grid_size[1:])
        enc2 = self.encoder2(proj_x2)
        x3 = hidden_states_out[1]
        proj_x3 = self.proj_feat(x3, self.hidden_size//self.grid_size[0], self.grid_size) # (batch, hidden_size, s, 2, 16, 16)
        proj_x3 = proj_x3.view(proj_x3.shape[0], -1, *self.grid_size[1:])
        enc3 = self.encoder3(proj_x3)
        
        proj_x = self.proj_feat(x, self.hidden_size//self.grid_size[0], self.grid_size)
        dec3 = proj_x.view(proj_x.shape[0], -1, *self.grid_size[1:])
        dec2 = self.decoder4(dec3, enc3)
        dec1 = self.decoder3(dec2, enc2)
        out = self.decoder2(dec1, enc1)
        seg_out = self.out(out)
        seg_pred = seg_out.view(seg_out.shape[0], -1, self.slice_num, *seg_out.shape[2:]) # (B, 4, slice, T, H, W)
        return seg_pred


class ImagingMaskedDecoder(nn.Module):
    def __init__(self, 
                grid_size: tuple,
                head_out_dim: int,
                decoder_num_patches: int,
                use_enc_pe: bool = True,
                enc_embed_dim: int = 1025,
                dec_embed_dim: int = 1025,
                dec_num_heads: int = 5,
                dec_depth: int = 2,
                mlp_ratio: float = 4.0,
                use_both_axes: bool = False,
                *args, **kwargs):
        super().__init__()
        self.grid_size = grid_size
        self.dec_embed_dim = dec_embed_dim
        self.use_both_axes = use_both_axes
        self.decoder_num_patches = decoder_num_patches
        self.use_enc_pe = use_enc_pe

        print(f"grid_size: {grid_size}, head_out_dim: {head_out_dim}, decoder_num_patches: {decoder_num_patches}, use_enc_pe: {use_enc_pe}, enc_embed_dim: {enc_embed_dim}, dec_embed_dim: {dec_embed_dim}, dec_num_heads: {dec_num_heads}, dec_depth: {dec_depth}, mlp_ratio: {mlp_ratio}, use_both_axes: {use_both_axes}")
        self.decoder_embed = nn.Linear(enc_embed_dim, dec_embed_dim)
        self.dec_pos_embed = nn.Parameter(
            torch.zeros(1, 1 + self.decoder_num_patches, dec_embed_dim), requires_grad=False)
        self.decoder = ViTDecoder(dim=dec_embed_dim, 
                                  num_heads=dec_num_heads,
                                  depth=dec_depth,
                                  mlp_ratio=mlp_ratio,)
        self.recon_head = nn.Linear(in_features=dec_embed_dim,
                                    out_features=head_out_dim,)
        self.mask_token = nn.Parameter(torch.zeros(1, 1, self.dec_embed_dim), requires_grad=True)
        self.initialize_parameters()
    
    def initialize_parameters(self):
        dec_pos_embed = sincos_pos_embed(self.dec_embed_dim, self.grid_size, cls_token=True,
                                         use_both_axes=self.use_both_axes)
        if not self.use_enc_pe:
            dec_pos_embed[:, 0] = 0
        self.dec_pos_embed.data.copy_(dec_pos_embed.unsqueeze(0))
        if hasattr(self, "mask_token"):
            torch.nn.init.normal_(self.mask_token, std=.02)

        # Initialize nn.Linear and nn.LayerNorm
        self.apply(self._init_weights)
        
    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward(self, x, ids_restore):
        """Forward pass of reconstruction decoder
        input:
            x: [B, 1 + length * mask_ratio, embed_dim] torch.Tensor
            ids_restore: [B, 1 + length * mask_ratio] torch.Tensor
        output:
            pred: [B, length, embed_dim] torch.Tensor
        """
        # Embed tokens
        x = self.decoder_embed(x)
        
        dec_pos_embed = self.dec_pos_embed.repeat(x.shape[0], 1, 1)
        
        # Append mask tokens and add positional embedding in schuffled order
        if ids_restore is not None:
            mask_token_n = ids_restore.shape[1] + 1 - x.shape[1]
            mask_tokens = self.mask_token.repeat(x.shape[0], mask_token_n, 1)
            x_shuffle = torch.cat([x[:, 1:, :], mask_tokens], dim=1)
            x_restore = torch.gather(x_shuffle, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, x_shuffle.shape[-1]))
            x_restore_ = x_restore + dec_pos_embed[:, 1:, :]
        else:
            x_restore_ = x[:, 1:, :] + dec_pos_embed[:, 1:, :]
        cls_tok_ = x[:, :1, :] + dec_pos_embed[:, :1, :]
        
        # Reconstruction decoder
        x = torch.cat([cls_tok_, x_restore_], dim=1) # add class token
        x = self.decoder(x) # apply transformer decoder
        
        # Reconstruction head
        x = self.recon_head(x)
        x = x[:, 1:, :] # remove cls token
        x = torch.sigmoid(x) # scale x to [0, 1] for reconstruction task
        return x


class ContrastiveLoss(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.criterion = torch.nn.CosineSimilarity(dim=1)
        
    # def forward(self, p1, p2, z1, z2):
    #     """y is detached from the computation of the gradient.
    #     This is inspired by paper: Chen and He, et al. 2020. Exploring simple siamese representation learning."""
    #     return -(self.criterion(p1, z2).mean() + self.criterion(p2, z1).mean()) * 0.5
    
    def forward(self, p2, z1):
        """y is detached from the computation of the gradient.
        This is inspired by paper: Chen and He, et al. 2020. Exploring simple siamese representation learning."""
        return -self.criterion(p2, z1).mean()

class ReconstructionCriterion(torch.nn.Module):
    def __init__(self, loss_types, loss_weights: Optional[list[float]] = None, mask_loss: bool = False, **kwargs):
        super().__init__()
        self.max_value = 1.0
        self.loss_types = loss_types
        self.loss_weights = loss_weights if loss_weights is not None else [1.0] * len(self.loss_types)
        self.loss_fcts = []
        loss_fct_dict = {"mse": torch.nn.MSELoss(reduction="none"), "cl": ContrastiveLoss()}
        if not isinstance(self.loss_types, list):
            self.loss_fcts = [loss_fct_dict[self.loss_types]]
        else:
            self.loss_fcts += [loss_fct_dict[loss_name] for loss_name in self.loss_types]

        self.use_mask_in_loss = mask_loss
        
    def forward(self, x, y, mask: Optional[torch.Tensor] = None, **kwargs):
        """Compute reconstruction loss
        x: [B, L, D] torch.Tensor, reconstructed image
        y: [B, L, D] torch.Tensor, reference image
        mask: [B, L] torch.Tensor, mask for encoder, where 0 is keep, 1 is remove
        
        loss_dict: dict, loss values
        psnr_value: float, psnr value
        """
        assert x.shape[-1] == y.shape[-1]
        if self.use_mask_in_loss:
            masked_x = x[mask == 1]
            masked_y = y[mask == 1]
        else:
            masked_x = x
            masked_y = y
        # --------------------------------------------------------------------------
        # Calculate losses
        total_loss = 0.0
        loss_dict = {}
        for i, loss_name in enumerate(self.loss_types):
            self.loss_fct = self.loss_fcts[i]
            if loss_name == "cl":
                p2, z1 = kwargs["p2"], kwargs["z1"]
                loss = self.loss_fct(p2, z1)
                # p1, p2, z1, z2 = kwargs["p1"], kwargs["p2"], kwargs["z1"], kwargs["z2"]
                # loss = self.loss_fct(p1, p2, z1, z2)
            else:    
                loss = self.loss_fct(masked_x, masked_y)
                loss = loss.mean()
            loss_dict[loss_name] = loss
            total_loss += self.loss_weights[i] * loss
        loss_dict["loss"] = total_loss
        return loss_dict


from dataclasses import dataclass, field
from typing import Any, Optional

import torch
import torch.nn as nn
from omegaconf import MISSING, OmegaConf

from ..nn.act import build_act
from ..nn.norm import build_norm
from ..nn.ops import (
    ChannelDuplicatingPixelUnshuffleUpSampleLayer,
    ConvLayer,
    ConvPixelShuffleUpSampleLayer,
    ConvPixelUnshuffleDownSampleLayer,
    EfficientViTBlock,
    IdentityLayer,
    InterpolateConvUpSampleLayer,
    OpSequential,
    PixelUnshuffleChannelAveragingDownSampleLayer,
    ResBlock,
    ResidualBlock,
)

__all__ = ["DCAE", "dc_ae_f32c32", "dc_ae_f64c128", "dc_ae_f128c512"]


@dataclass
class EncoderConfig:
    in_channels: int = MISSING
    latent_channels: int = MISSING
    width_list: tuple[int, ...] = (128, 256, 512, 512, 1024, 1024)
    depth_list: tuple[int, ...] = (2, 2, 2, 2, 2, 2)
    block_type: Any = "ResBlock"
    norm: str = "trms2d"
    act: str = "silu"
    downsample_block_type: str = "ConvPixelUnshuffle"
    downsample_match_channel: bool = True
    downsample_shortcut: Optional[str] = "averaging"
    out_norm: Optional[str] = None
    out_act: Optional[str] = None
    out_shortcut: Optional[str] = "averaging"
    double_latent: bool = False


@dataclass
class DecoderConfig:
    in_channels: int = MISSING
    latent_channels: int = MISSING
    in_shortcut: Optional[str] = "duplicating"
    width_list: tuple[int, ...] = (128, 256, 512, 512, 1024, 1024)
    depth_list: tuple[int, ...] = (2, 2, 2, 2, 2, 2)
    block_type: Any = "ResBlock"
    norm: Any = "trms2d"
    act: Any = "silu"
    upsample_block_type: str = "ConvPixelShuffle"
    upsample_match_channel: bool = True
    upsample_shortcut: str = "duplicating"
    out_norm: str = "trms2d"
    out_act: str = "relu"


@dataclass
class DCAEConfig:
    in_channels: int = 3
    latent_channels: int = 32
    encoder: EncoderConfig = field(
        default_factory=lambda: EncoderConfig(in_channels="${..in_channels}", latent_channels="${..latent_channels}")
    )
    decoder: DecoderConfig = field(
        default_factory=lambda: DecoderConfig(in_channels="${..in_channels}", latent_channels="${..latent_channels}")
    )
    use_quant_conv: bool = False

    pretrained_path: Optional[str] = None
    pretrained_source: str = "dc-ae"

    scaling_factor: Optional[float] = None


def build_block(
    block_type: str, in_channels: int, out_channels: int, norm: Optional[str], act: Optional[str]
) -> nn.Module:
    if block_type == "ResBlock":
        assert in_channels == out_channels
        main_block = ResBlock(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            use_bias=(True, False),
            norm=(None, norm),
            act_func=(act, None),
        )
        block = ResidualBlock(main_block, IdentityLayer())
    elif block_type == "EViT_GLU":
        assert in_channels == out_channels
        block = EfficientViTBlock(in_channels, norm=norm, act_func=act, local_module="GLUMBConv", scales=())
    elif block_type == "EViTS5_GLU":
        assert in_channels == out_channels
        block = EfficientViTBlock(in_channels, norm=norm, act_func=act, local_module="GLUMBConv", scales=(5,))
    else:
        raise ValueError(f"block_type {block_type} is not supported")
    return block


def build_stage_main(
    width: int, depth: int, block_type: str | list[str], norm: str, act: str, input_width: int
) -> list[nn.Module]:
    assert isinstance(block_type, str) or (isinstance(block_type, list) and depth == len(block_type))
    stage = []
    for d in range(depth):
        current_block_type = block_type[d] if isinstance(block_type, list) else block_type
        block = build_block(
            block_type=current_block_type,
            in_channels=width if d > 0 else input_width,
            out_channels=width,
            norm=norm,
            act=act,
        )
        stage.append(block)
    return stage


def build_downsample_block(block_type: str, in_channels: int, out_channels: int, shortcut: Optional[str]) -> nn.Module:
    if block_type == "Conv":
        block = ConvLayer(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=2,
            use_bias=True,
            norm=None,
            act_func=None,
        )
    elif block_type == "ConvPixelUnshuffle":
        block = ConvPixelUnshuffleDownSampleLayer(
            in_channels=in_channels, out_channels=out_channels, kernel_size=3, factor=2
        )
    else:
        raise ValueError(f"block_type {block_type} is not supported for downsampling")
    if shortcut is None:
        pass
    elif shortcut == "averaging":
        shortcut_block = PixelUnshuffleChannelAveragingDownSampleLayer(
            in_channels=in_channels, out_channels=out_channels, factor=2
        )
        block = ResidualBlock(block, shortcut_block)
    else:
        raise ValueError(f"shortcut {shortcut} is not supported for downsample")
    return block


def build_upsample_block(block_type: str, in_channels: int, out_channels: int, shortcut: Optional[str]) -> nn.Module:
    if block_type == "ConvPixelShuffle":
        block = ConvPixelShuffleUpSampleLayer(
            in_channels=in_channels, out_channels=out_channels, kernel_size=3, factor=2
        )
    elif block_type == "InterpolateConv":
        block = InterpolateConvUpSampleLayer(
            in_channels=in_channels, out_channels=out_channels, kernel_size=3, factor=2
        )
    else:
        raise ValueError(f"block_type {block_type} is not supported for upsampling")
    if shortcut is None:
        pass
    elif shortcut == "duplicating":
        shortcut_block = ChannelDuplicatingPixelUnshuffleUpSampleLayer(
            in_channels=in_channels, out_channels=out_channels, factor=2
        )
        block = ResidualBlock(block, shortcut_block)
    else:
        raise ValueError(f"shortcut {shortcut} is not supported for upsample")
    return block


def build_encoder_project_in_block(in_channels: int, out_channels: int, factor: int, downsample_block_type: str):
    if factor == 1:
        block = ConvLayer(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            use_bias=True,
            norm=None,
            act_func=None,
        )
    elif factor == 2:
        block = build_downsample_block(
            block_type=downsample_block_type, in_channels=in_channels, out_channels=out_channels, shortcut=None
        )
    else:
        raise ValueError(f"downsample factor {factor} is not supported for encoder project in")
    return block


def build_encoder_project_out_block(
    in_channels: int, out_channels: int, norm: Optional[str], act: Optional[str], shortcut: Optional[str]
):
    block = OpSequential(
        [
            build_norm(norm),
            build_act(act),
            ConvLayer(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=3,
                stride=1,
                use_bias=True,
                norm=None,
                act_func=None,
            ),
        ]
    )
    if shortcut is None:
        pass
    elif shortcut == "averaging":
        shortcut_block = PixelUnshuffleChannelAveragingDownSampleLayer(
            in_channels=in_channels, out_channels=out_channels, factor=1
        )
        block = ResidualBlock(block, shortcut_block)
    else:
        raise ValueError(f"shortcut {shortcut} is not supported for encoder project out")
    return block


def build_decoder_project_in_block(in_channels: int, out_channels: int, shortcut: Optional[str]):
    block = ConvLayer(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=3,
        stride=1,
        use_bias=True,
        norm=None,
        act_func=None,
    )
    if shortcut is None:
        pass
    elif shortcut == "duplicating":
        shortcut_block = ChannelDuplicatingPixelUnshuffleUpSampleLayer(
            in_channels=in_channels, out_channels=out_channels, factor=1
        )
        block = ResidualBlock(block, shortcut_block)
    else:
        raise ValueError(f"shortcut {shortcut} is not supported for decoder project in")
    return block


def build_decoder_project_out_block(
    in_channels: int, out_channels: int, factor: int, upsample_block_type: str, norm: Optional[str], act: Optional[str]
):
    layers: list[nn.Module] = [
        build_norm(norm, in_channels),
        build_act(act),
    ]
    if factor == 1:
        layers.append(
            ConvLayer(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=3,
                stride=1,
                use_bias=True,
                norm=None,
                act_func=None,
            )
        )
    elif factor == 2:
        layers.append(
            build_upsample_block(
                block_type=upsample_block_type, in_channels=in_channels, out_channels=out_channels, shortcut=None
            )
        )
    else:
        raise ValueError(f"upsample factor {factor} is not supported for decoder project out")
    return OpSequential(layers)


class Encoder(nn.Module):
    def __init__(self, cfg: EncoderConfig):
        super().__init__()
        self.cfg = cfg
        num_stages = len(cfg.width_list)
        self.num_stages = num_stages
        assert len(cfg.depth_list) == num_stages
        assert len(cfg.width_list) == num_stages
        assert isinstance(cfg.block_type, str) or (
            isinstance(cfg.block_type, list) and len(cfg.block_type) == num_stages
        )

        self.project_in = build_encoder_project_in_block(
            in_channels=cfg.in_channels,
            out_channels=cfg.width_list[0] if cfg.depth_list[0] > 0 else cfg.width_list[1],
            factor=1 if cfg.depth_list[0] > 0 else 2,
            downsample_block_type=cfg.downsample_block_type,
        )

        self.stages: list[OpSequential] = []
        for stage_id, (width, depth) in enumerate(zip(cfg.width_list, cfg.depth_list)):
            block_type = cfg.block_type[stage_id] if isinstance(cfg.block_type, list) else cfg.block_type
            stage = build_stage_main(
                width=width, depth=depth, block_type=block_type, norm=cfg.norm, act=cfg.act, input_width=width
            )

            if stage_id < num_stages - 1 and depth > 0:
                downsample_block = build_downsample_block(
                    block_type=cfg.downsample_block_type,
                    in_channels=width,
                    out_channels=cfg.width_list[stage_id + 1] if cfg.downsample_match_channel else width,
                    shortcut=cfg.downsample_shortcut,
                )
                stage.append(downsample_block)
            self.stages.append(OpSequential(stage))
        self.stages = nn.ModuleList(self.stages)

        self.project_out = build_encoder_project_out_block(
            in_channels=cfg.width_list[-1],
            out_channels=2 * cfg.latent_channels if cfg.double_latent else cfg.latent_channels,
            norm=cfg.out_norm,
            act=cfg.out_act,
            shortcut=cfg.out_shortcut,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.project_in(x)
        for stage in self.stages:
            if len(stage.op_list) == 0:
                continue
            x = stage(x)
        x = self.project_out(x)
        return x


class Decoder(nn.Module):
    def __init__(self, cfg: DecoderConfig):
        super().__init__()
        self.cfg = cfg
        num_stages = len(cfg.width_list)
        self.num_stages = num_stages
        assert len(cfg.depth_list) == num_stages
        assert len(cfg.width_list) == num_stages
        assert isinstance(cfg.block_type, str) or (
            isinstance(cfg.block_type, list) and len(cfg.block_type) == num_stages
        )
        assert isinstance(cfg.norm, str) or (isinstance(cfg.norm, list) and len(cfg.norm) == num_stages)
        assert isinstance(cfg.act, str) or (isinstance(cfg.act, list) and len(cfg.act) == num_stages)

        self.project_in = build_decoder_project_in_block(
            in_channels=cfg.latent_channels,
            out_channels=cfg.width_list[-1],
            shortcut=cfg.in_shortcut,
        )

        self.stages: list[OpSequential] = []
        for stage_id, (width, depth) in reversed(list(enumerate(zip(cfg.width_list, cfg.depth_list)))):
            stage = []
            if stage_id < num_stages - 1 and depth > 0:
                upsample_block = build_upsample_block(
                    block_type=cfg.upsample_block_type,
                    in_channels=cfg.width_list[stage_id + 1],
                    out_channels=width if cfg.upsample_match_channel else cfg.width_list[stage_id + 1],
                    shortcut=cfg.upsample_shortcut,
                )
                stage.append(upsample_block)

            block_type = cfg.block_type[stage_id] if isinstance(cfg.block_type, list) else cfg.block_type
            norm = cfg.norm[stage_id] if isinstance(cfg.norm, list) else cfg.norm
            act = cfg.act[stage_id] if isinstance(cfg.act, list) else cfg.act
            stage.extend(
                build_stage_main(
                    width=width,
                    depth=depth,
                    block_type=block_type,
                    norm=norm,
                    act=act,
                    input_width=(
                        width if cfg.upsample_match_channel else cfg.width_list[min(stage_id + 1, num_stages - 1)]
                    ),
                )
            )
            self.stages.insert(0, OpSequential(stage))
        self.stages = nn.ModuleList(self.stages)

        self.project_out = build_decoder_project_out_block(
            in_channels=cfg.width_list[0] if cfg.depth_list[0] > 0 else cfg.width_list[1],
            out_channels=cfg.in_channels,
            factor=1 if cfg.depth_list[0] > 0 else 2,
            upsample_block_type=cfg.upsample_block_type,
            norm=cfg.out_norm,
            act=cfg.out_act,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.project_in(x)
        for stage in reversed(self.stages):
            if len(stage.op_list) == 0:
                continue
            x = stage(x)
        x = self.project_out(x)
        return x


class DCAE(nn.Module):
    def __init__(self, cfg: DCAEConfig):
        super().__init__()
        self.cfg = cfg
        self.encoder = Encoder(cfg.encoder)
        self.decoder = Decoder(cfg.decoder)

        if self.cfg.pretrained_path is not None:
            self.load_model()

    def load_model(self):
        if self.cfg.pretrained_source == "dc-ae":
            state_dict = torch.load(self.cfg.pretrained_path, map_location="cpu", weights_only=True)["state_dict"]
            self.load_state_dict(state_dict)
        else:
            raise NotImplementedError

    @property
    def spatial_compression_ratio(self) -> int:
        return 2 ** (self.decoder.num_stages - 1)

    def encode(self, x: torch.Tensor) -> torch.Tensor:
        x = self.encoder(x)
        return x

    def decode(self, x: torch.Tensor) -> torch.Tensor:
        x = self.decoder(x)
        return x

    def forward(self, x: torch.Tensor, global_step: int) -> torch.Tensor:
        x = self.encoder(x)
        x = self.decoder(x)
        return x, torch.tensor(0), {}


def dc_ae_f32c32(name: str, pretrained_path: str) -> DCAEConfig:
    if name in ["dc-ae-f32c32-in-1.0", "dc-ae-f32c32-mix-1.0"]:
        cfg_str = (
            "latent_channels=32 "
            "encoder.block_type=[ResBlock,ResBlock,ResBlock,EViT_GLU,EViT_GLU,EViT_GLU] "
            "encoder.width_list=[128,256,512,512,1024,1024] encoder.depth_list=[0,4,8,2,2,2] "
            "decoder.block_type=[ResBlock,ResBlock,ResBlock,EViT_GLU,EViT_GLU,EViT_GLU] "
            "decoder.width_list=[128,256,512,512,1024,1024] decoder.depth_list=[0,5,10,2,2,2] "
            "decoder.norm=[bn2d,bn2d,bn2d,trms2d,trms2d,trms2d] decoder.act=[relu,relu,relu,silu,silu,silu]"
        )
    elif name in ["dc-ae-f32c32-sana-1.0"]:
        cfg_str = (
            "latent_channels=32 "
            "encoder.block_type=[ResBlock,ResBlock,ResBlock,EViTS5_GLU,EViTS5_GLU,EViTS5_GLU] "
            "encoder.width_list=[128,256,512,512,1024,1024] encoder.depth_list=[2,2,2,3,3,3] "
            "encoder.downsample_block_type=Conv "
            "decoder.block_type=[ResBlock,ResBlock,ResBlock,EViTS5_GLU,EViTS5_GLU,EViTS5_GLU] "
            "decoder.width_list=[128,256,512,512,1024,1024] decoder.depth_list=[3,3,3,3,3,3] "
            "decoder.upsample_block_type=InterpolateConv "
            "decoder.norm=trms2d decoder.act=silu "
            "scaling_factor=0.41407"
        )
    else:
        raise NotImplementedError
    cfg = OmegaConf.from_dotlist(cfg_str.split(" "))
    cfg: DCAEConfig = OmegaConf.to_object(OmegaConf.merge(OmegaConf.structured(DCAEConfig), cfg))
    cfg.pretrained_path = pretrained_path
    return cfg


def dc_ae_f64c128(name: str, pretrained_path: Optional[str] = None) -> DCAEConfig:
    if name in ["dc-ae-f64c128-in-1.0", "dc-ae-f64c128-mix-1.0"]:
        cfg_str = (
            "latent_channels=128 "
            "encoder.block_type=[ResBlock,ResBlock,ResBlock,EViT_GLU,EViT_GLU,EViT_GLU,EViT_GLU] "
            "encoder.width_list=[128,256,512,512,1024,1024,2048] encoder.depth_list=[0,4,8,2,2,2,2] "
            "decoder.block_type=[ResBlock,ResBlock,ResBlock,EViT_GLU,EViT_GLU,EViT_GLU,EViT_GLU] "
            "decoder.width_list=[128,256,512,512,1024,1024,2048] decoder.depth_list=[0,5,10,2,2,2,2] "
            "decoder.norm=[bn2d,bn2d,bn2d,trms2d,trms2d,trms2d,trms2d] decoder.act=[relu,relu,relu,silu,silu,silu,silu]"
        )
    else:
        raise NotImplementedError
    cfg = OmegaConf.from_dotlist(cfg_str.split(" "))
    cfg: DCAEConfig = OmegaConf.to_object(OmegaConf.merge(OmegaConf.structured(DCAEConfig), cfg))
    cfg.pretrained_path = pretrained_path
    return cfg


def dc_ae_f128c512(name: str, pretrained_path: Optional[str] = None) -> DCAEConfig:
    if name in ["dc-ae-f128c512-in-1.0", "dc-ae-f128c512-mix-1.0"]:
        cfg_str = (
            "latent_channels=512 "
            "encoder.block_type=[ResBlock,ResBlock,ResBlock,EViT_GLU,EViT_GLU,EViT_GLU,EViT_GLU,EViT_GLU] "
            "encoder.width_list=[128,256,512,512,1024,1024,2048,2048] encoder.depth_list=[0,4,8,2,2,2,2,2] "
            "decoder.block_type=[ResBlock,ResBlock,ResBlock,EViT_GLU,EViT_GLU,EViT_GLU,EViT_GLU,EViT_GLU] "
            "decoder.width_list=[128,256,512,512,1024,1024,2048,2048] decoder.depth_list=[0,5,10,2,2,2,2,2] "
            "decoder.norm=[bn2d,bn2d,bn2d,trms2d,trms2d,trms2d,trms2d,trms2d] decoder.act=[relu,relu,relu,silu,silu,silu,silu,silu]"
        )
    else:
        raise NotImplementedError
    cfg = OmegaConf.from_dotlist(cfg_str.split(" "))
    cfg: DCAEConfig = OmegaConf.to_object(OmegaConf.merge(OmegaConf.structured(DCAEConfig), cfg))
    cfg.pretrained_path = pretrained_path
    return cfg

if __name__ == "__main__":
    cfg = dc_ae_f128c512("dc-ae-f128c512-in-1.0", pretrained_path="pretrained/dc-ae-f128c512-in-1.0.pth")
    model = DCAE(cfg)
    print(model)


import torch
import torch.nn as nn
from typing import Dict, Any, Optional
from medlat.utils import init_from_ckpt, get_model_type


class GenWrapper(nn.Module):
    def __init__(
        self,
        generator: Optional[Dict[str, Any]],
        first_stage: Optional[Dict[str, Any]],
        scale_factor: float = None,
        ckpt_path: Optional[str] = None,
        scale_steps: int = 100,
    ):
        super().__init__()
        self.generator = generator
        self.generator_type = get_model_type(generator)
        self.first_stage = first_stage
        self.first_stage_type = get_model_type(first_stage)
        print(f"generator_type: {self.generator_type}, first_stage_type: {self.first_stage_type}")
        self.scale_steps = scale_steps

        if self.first_stage_type == "discrete" and self.generator_type == "non-autoregressive":
            self.fcn_encode = self.first_stage.encode_to_prequant
            self.fcn_decode = self.first_stage.decode_from_prequant

        elif self.first_stage_type == "continuous" and self.generator_type == "non-autoregressive":
            self.fcn_encode = self.first_stage.encode
            self.fcn_decode = self.first_stage.decode

        elif self.first_stage_type == "discrete" and self.generator_type == "autoregressive":
            self.fcn_encode = self.first_stage.encode
            self.fcn_decode = self.first_stage.decode_code

        elif self.first_stage_type == "continuous" and self.generator_type == "autoregressive":
            self.fcn_encode = self.first_stage.encode
            self.fcn_decode = self.first_stage.decode

        else:
            raise ValueError(f"Unsupported combination of generator and first stage types: {self.generator_type} and {self.first_stage_type}")

        # Determine if we should do automatic scale_factor estimation
        self._auto_scale_factor = scale_factor is None
        
        # Register scale factor as buffer
        # If None, initialize to 1.0 (will be updated automatically)
        # If provided, use the given value (will not be updated)
        initial_scale = 1.0 if scale_factor is None else scale_factor
        self.register_buffer("scale_factor", torch.tensor(initial_scale))
        
        # Track step counter and running statistics for automatic scale_factor determination
        # Only used when _auto_scale_factor is True
        self._scale_step_counter = 0
        self._running_std_sum = 0.0

        if ckpt_path is not None:
            init_from_ckpt(self, ckpt_path)

        # Freeze first stage
        if self.first_stage is not None:
            for p in self.first_stage.parameters():
                p.requires_grad = False
            self.first_stage.eval()

    # ---------------------------------------------------------------------
    # Training mode handling
    # ---------------------------------------------------------------------
    def train(self, mode: bool = True):
        super().train(mode)
        if self.generator is not None:
            self.generator.train(mode)

        if self.first_stage is not None:
            self.first_stage.eval()
            for p in self.first_stage.parameters():
                p.requires_grad = False
        return self

    # ---------------------------------------------------------------------
    # Scale factor determination
    # ---------------------------------------------------------------------
    def _update_scale_factor(self, quant: torch.Tensor) -> None:
        """
        Automatically determine scale_factor during the first scale_steps steps.
        Updates scale_factor based on the standard deviation of quantized latents.
        After scale_steps steps, the scale_factor remains fixed.
        Only updates if scale_factor was initially None.
        """
        # Only do automatic estimation if scale_factor was None
        if not self._auto_scale_factor:
            return
        
        if self._scale_step_counter > self.scale_steps:
            return
        
        if self._scale_step_counter == self.scale_steps:
            print(f"Scale factor fixed at {self.scale_factor.item()}")
            self._scale_step_counter +=1
            return
        
        with torch.no_grad():
            # Compute standard deviation of quantized latents
            quant_std = quant.std().item()
            
            # Accumulate std values
            self._running_std_sum += quant_std
            self._scale_step_counter += 1
            
            # Compute average std and update scale_factor
            avg_std = self._running_std_sum / self._scale_step_counter
            self.scale_factor.data = torch.tensor(1.0 / (avg_std + 1e-8), device=self.scale_factor.device)

    # ---------------------------------------------------------------------
    # Encoding
    # ---------------------------------------------------------------------
    def vae_encode(self, image: torch.Tensor, sample: bool = True) -> torch.Tensor:
        if self.first_stage is None:
            return image
        
        quant, loss, info = self.fcn_encode(image)
        self.quant_shape = quant.permute(0, 2, 3, 1).shape  # (B, H, W, C) for indices decoding later


        # Automatically determine scale_factor during the first scale_steps steps
        if self.training:
            self._update_scale_factor(quant)
        quant = quant * self.scale_factor

        # ---- autoregressive path ----
        if info is None:
            return quant
        else:
            _, _, indices = info
            if type(indices) == torch.Tensor: ## normal
                return indices.reshape(image.shape[0], -1)
            elif isinstance(indices, (list, tuple)):  # residual quantizer
                indices = [ind.reshape(image.shape[0], -1) for ind in indices]
                return torch.cat(indices, dim=1)
            return quant


    # ---------------------------------------------------------------------
    # Decoding
    # ---------------------------------------------------------------------
    def vae_decode(self, z: torch.Tensor, out_shape = None) -> torch.Tensor:
        if self.first_stage is None:
            return z
        
        if self.generator_type == "autoregressive" and self.first_stage_type == "discrete":
            return self.fcn_decode(z, out_shape=out_shape if out_shape is not None else self.quant_shape)
        else:
            return self.fcn_decode(z / self.scale_factor)

    # ---------------------------------------------------------------------
    # Forward
    # ---------------------------------------------------------------------
    def forward(self, x, *args, **kwargs):
        return self.generator.forward(x, *args, **kwargs)



import torch
import numpy as np


class _DeterministicPosterior:
    """Posterior when encoder outputs z directly (double_z=False, no Gaussian)."""

    def __init__(self, z: torch.Tensor):
        self.z = z

    def sample(self) -> torch.Tensor:
        return self.z

    def mode(self) -> torch.Tensor:
        return self.z

    def kl(self) -> torch.Tensor:
        return torch.zeros((), device=self.z.device, dtype=self.z.dtype)

class DiagonalGaussianDistribution(object):
    def __init__(self, parameters, deterministic=False, channel_dim=1):
        self.parameters = parameters
        self.mean, self.logvar = torch.chunk(parameters, 2, dim=channel_dim)
        self.logvar = torch.clamp(self.logvar, -30.0, 20.0)
        self.deterministic = deterministic
        self.std = torch.exp(0.5 * self.logvar)
        self.var = torch.exp(self.logvar)
        if self.deterministic:
            self.var = self.std = torch.zeros_like(self.mean).to(
                device=self.parameters.device
            )

    def sample(self):
        x = self.mean + self.std * torch.randn(self.mean.shape).to(
            device=self.parameters.device
        )
        return x

    def kl(self, other=None):
        if self.deterministic:
            return torch.tensor([0.0], device=self.mean.device)
        else:
            reduce_dims = list(range(1, self.mean.ndim))  # all dims except batch
            if other is None:
                return 0.5 * torch.sum(
                    torch.pow(self.mean, 2) + self.var - 1.0 - self.logvar,
                    dim=reduce_dims
                )
            else:
                return 0.5 * torch.sum(
                    torch.pow(self.mean - other.mean, 2) / other.var
                    + self.var / other.var
                    - 1.0
                    - self.logvar
                    + other.logvar,
                    dim=reduce_dims
                )

    def nll(self, sample):
        if self.deterministic:
            return torch.tensor([0.0], device=self.mean.device)
        logtwopi = np.log(2.0 * np.pi)
        reduce_dims = list(range(1, self.mean.ndim))  # sum over everything but batch
        return 0.5 * torch.sum(
            logtwopi + self.logvar + torch.pow(sample - self.mean, 2) / self.var,
            dim=reduce_dims
        )

    def mode(self):
        return self.mean
    


from typing import Optional

import torch
import torch.nn as nn

from ..nn import (
    ConvLayer,
    DSConv,
    EfficientViTBlock,
    FusedMBConv,
    IdentityLayer,
    MBConv,
    OpSequential,
    ResBlock,
    ResidualBlock,
)
from ..utils import build_kwargs_from_config

__all__ = [
    "EfficientViTBackbone",
    "efficientvit_backbone_b0",
    "efficientvit_backbone_b1",
    "efficientvit_backbone_b2",
    "efficientvit_backbone_b3",
    "EfficientViTLargeBackbone",
    "efficientvit_backbone_l0",
    "efficientvit_backbone_l1",
    "efficientvit_backbone_l2",
    "efficientvit_backbone_l3",
]


class EfficientViTBackbone(nn.Module):
    def __init__(
        self,
        width_list: list[int],
        depth_list: list[int],
        in_channels=3,
        dim=32,
        expand_ratio=4,
        norm="bn2d",
        act_func="hswish",
    ) -> None:
        super().__init__()

        self.width_list = []
        # input stem
        self.input_stem = [
            ConvLayer(
                in_channels=in_channels,
                out_channels=width_list[0],
                stride=2,
                norm=norm,
                act_func=act_func,
            )
        ]
        for _ in range(depth_list[0]):
            block = self.build_local_block(
                in_channels=width_list[0],
                out_channels=width_list[0],
                stride=1,
                expand_ratio=1,
                norm=norm,
                act_func=act_func,
            )
            self.input_stem.append(ResidualBlock(block, IdentityLayer()))
        in_channels = width_list[0]
        self.input_stem = OpSequential(self.input_stem)
        self.width_list.append(in_channels)

        # stages
        self.stages = []
        for w, d in zip(width_list[1:3], depth_list[1:3]):
            stage = []
            for i in range(d):
                stride = 2 if i == 0 else 1
                block = self.build_local_block(
                    in_channels=in_channels,
                    out_channels=w,
                    stride=stride,
                    expand_ratio=expand_ratio,
                    norm=norm,
                    act_func=act_func,
                )
                block = ResidualBlock(block, IdentityLayer() if stride == 1 else None)
                stage.append(block)
                in_channels = w
            self.stages.append(OpSequential(stage))
            self.width_list.append(in_channels)

        for w, d in zip(width_list[3:], depth_list[3:]):
            stage = []
            block = self.build_local_block(
                in_channels=in_channels,
                out_channels=w,
                stride=2,
                expand_ratio=expand_ratio,
                norm=norm,
                act_func=act_func,
                fewer_norm=True,
            )
            stage.append(ResidualBlock(block, None))
            in_channels = w

            for _ in range(d):
                stage.append(
                    EfficientViTBlock(
                        in_channels=in_channels,
                        dim=dim,
                        expand_ratio=expand_ratio,
                        norm=norm,
                        act_func=act_func,
                    )
                )
            self.stages.append(OpSequential(stage))
            self.width_list.append(in_channels)
        self.stages = nn.ModuleList(self.stages)

    @staticmethod
    def build_local_block(
        in_channels: int,
        out_channels: int,
        stride: int,
        expand_ratio: float,
        norm: str,
        act_func: str,
        fewer_norm: bool = False,
    ) -> nn.Module:
        if expand_ratio == 1:
            block = DSConv(
                in_channels=in_channels,
                out_channels=out_channels,
                stride=stride,
                use_bias=(True, False) if fewer_norm else False,
                norm=(None, norm) if fewer_norm else norm,
                act_func=(act_func, None),
            )
        else:
            block = MBConv(
                in_channels=in_channels,
                out_channels=out_channels,
                stride=stride,
                expand_ratio=expand_ratio,
                use_bias=(True, True, False) if fewer_norm else False,
                norm=(None, None, norm) if fewer_norm else norm,
                act_func=(act_func, act_func, None),
            )
        return block

    def forward(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        output_dict = {"input": x}
        output_dict["stage0"] = x = self.input_stem(x)
        for stage_id, stage in enumerate(self.stages, 1):
            output_dict["stage%d" % stage_id] = x = stage(x)
        output_dict["stage_final"] = x
        return output_dict


def efficientvit_backbone_b0(**kwargs) -> EfficientViTBackbone:
    backbone = EfficientViTBackbone(
        width_list=[8, 16, 32, 64, 128],
        depth_list=[1, 2, 2, 2, 2],
        dim=16,
        **build_kwargs_from_config(kwargs, EfficientViTBackbone),
    )
    return backbone


def efficientvit_backbone_b1(**kwargs) -> EfficientViTBackbone:
    backbone = EfficientViTBackbone(
        width_list=[16, 32, 64, 128, 256],
        depth_list=[1, 2, 3, 3, 4],
        dim=16,
        **build_kwargs_from_config(kwargs, EfficientViTBackbone),
    )
    return backbone


def efficientvit_backbone_b2(**kwargs) -> EfficientViTBackbone:
    backbone = EfficientViTBackbone(
        width_list=[24, 48, 96, 192, 384],
        depth_list=[1, 3, 4, 4, 6],
        dim=32,
        **build_kwargs_from_config(kwargs, EfficientViTBackbone),
    )
    return backbone


def efficientvit_backbone_b3(**kwargs) -> EfficientViTBackbone:
    backbone = EfficientViTBackbone(
        width_list=[32, 64, 128, 256, 512],
        depth_list=[1, 4, 6, 6, 9],
        dim=32,
        **build_kwargs_from_config(kwargs, EfficientViTBackbone),
    )
    return backbone


class EfficientViTLargeBackbone(nn.Module):
    def __init__(
        self,
        width_list: list[int],
        depth_list: list[int],
        block_list: Optional[list[str]] = None,
        expand_list: Optional[list[float]] = None,
        fewer_norm_list: Optional[list[bool]] = None,
        in_channels=3,
        qkv_dim=32,
        norm="bn2d",
        act_func="gelu",
    ) -> None:
        super().__init__()
        block_list = ["res", "fmb", "fmb", "mb", "att"] if block_list is None else block_list
        expand_list = [1, 4, 4, 4, 6] if expand_list is None else expand_list
        fewer_norm_list = [False, False, False, True, True] if fewer_norm_list is None else fewer_norm_list

        self.width_list = []
        self.stages = []
        # stage 0
        stage0 = [
            ConvLayer(
                in_channels=in_channels,
                out_channels=width_list[0],
                stride=2,
                norm=norm,
                act_func=act_func,
            )
        ]
        for _ in range(depth_list[0]):
            block = self.build_local_block(
                block=block_list[0],
                in_channels=width_list[0],
                out_channels=width_list[0],
                stride=1,
                expand_ratio=expand_list[0],
                norm=norm,
                act_func=act_func,
                fewer_norm=fewer_norm_list[0],
            )
            stage0.append(ResidualBlock(block, IdentityLayer()))
        in_channels = width_list[0]
        self.stages.append(OpSequential(stage0))
        self.width_list.append(in_channels)

        for stage_id, (w, d) in enumerate(zip(width_list[1:], depth_list[1:]), start=1):
            stage = []
            block = self.build_local_block(
                block="mb" if block_list[stage_id] not in ["mb", "fmb"] else block_list[stage_id],
                in_channels=in_channels,
                out_channels=w,
                stride=2,
                expand_ratio=expand_list[stage_id] * 4,
                norm=norm,
                act_func=act_func,
                fewer_norm=fewer_norm_list[stage_id],
            )
            stage.append(ResidualBlock(block, None))
            in_channels = w

            for _ in range(d):
                if block_list[stage_id].startswith("att"):
                    stage.append(
                        EfficientViTBlock(
                            in_channels=in_channels,
                            dim=qkv_dim,
                            expand_ratio=expand_list[stage_id],
                            scales=(3,) if block_list[stage_id] == "att@3" else (5,),
                            norm=norm,
                            act_func=act_func,
                        )
                    )
                else:
                    block = self.build_local_block(
                        block=block_list[stage_id],
                        in_channels=in_channels,
                        out_channels=in_channels,
                        stride=1,
                        expand_ratio=expand_list[stage_id],
                        norm=norm,
                        act_func=act_func,
                        fewer_norm=fewer_norm_list[stage_id],
                    )
                    block = ResidualBlock(block, IdentityLayer())
                    stage.append(block)
            self.stages.append(OpSequential(stage))
            self.width_list.append(in_channels)
        self.stages = nn.ModuleList(self.stages)

    @staticmethod
    def build_local_block(
        block: str,
        in_channels: int,
        out_channels: int,
        stride: int,
        expand_ratio: float,
        norm: str,
        act_func: str,
        fewer_norm: bool = False,
    ) -> nn.Module:
        if block == "res":
            block = ResBlock(
                in_channels=in_channels,
                out_channels=out_channels,
                stride=stride,
                use_bias=(True, False) if fewer_norm else False,
                norm=(None, norm) if fewer_norm else norm,
                act_func=(act_func, None),
            )
        elif block == "fmb":
            block = FusedMBConv(
                in_channels=in_channels,
                out_channels=out_channels,
                stride=stride,
                expand_ratio=expand_ratio,
                use_bias=(True, False) if fewer_norm else False,
                norm=(None, norm) if fewer_norm else norm,
                act_func=(act_func, None),
            )
        elif block == "mb":
            block = MBConv(
                in_channels=in_channels,
                out_channels=out_channels,
                stride=stride,
                expand_ratio=expand_ratio,
                use_bias=(True, True, False) if fewer_norm else False,
                norm=(None, None, norm) if fewer_norm else norm,
                act_func=(act_func, act_func, None),
            )
        else:
            raise ValueError(block)
        return block

    def forward(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        output_dict = {"input": x}
        for stage_id, stage in enumerate(self.stages):
            output_dict["stage%d" % stage_id] = x = stage(x)
        output_dict["stage_final"] = x
        return output_dict


def efficientvit_backbone_l0(**kwargs) -> EfficientViTLargeBackbone:
    backbone = EfficientViTLargeBackbone(
        width_list=[32, 64, 128, 256, 512],
        depth_list=[1, 1, 1, 4, 4],
        **build_kwargs_from_config(kwargs, EfficientViTLargeBackbone),
    )
    return backbone


def efficientvit_backbone_l1(**kwargs) -> EfficientViTLargeBackbone:
    backbone = EfficientViTLargeBackbone(
        width_list=[32, 64, 128, 256, 512],
        depth_list=[1, 1, 1, 6, 6],
        **build_kwargs_from_config(kwargs, EfficientViTLargeBackbone),
    )
    return backbone


def efficientvit_backbone_l2(**kwargs) -> EfficientViTLargeBackbone:
    backbone = EfficientViTLargeBackbone(
        width_list=[32, 64, 128, 256, 512],
        depth_list=[1, 2, 2, 8, 8],
        **build_kwargs_from_config(kwargs, EfficientViTLargeBackbone),
    )
    return backbone


def efficientvit_backbone_l3(**kwargs) -> EfficientViTLargeBackbone:
    backbone = EfficientViTLargeBackbone(
        width_list=[64, 128, 256, 512, 1024],
        depth_list=[1, 2, 2, 8, 8],
        **build_kwargs_from_config(kwargs, EfficientViTLargeBackbone),
    )
    return backbone



from typing import Optional

import torch
import torch.nn as nn
from torch.nn.modules.batchnorm import _BatchNorm

from .triton_rms_norm import TritonRMSNorm2dFunc
from ..utils.network import build_kwargs_from_config

__all__ = ["LayerNorm2d", "TritonRMSNorm2d", "build_norm", "reset_bn", "set_norm_eps"]


class LayerNorm2d(nn.LayerNorm):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = x - torch.mean(x, dim=1, keepdim=True)
        out = out / torch.sqrt(torch.square(out).mean(dim=1, keepdim=True) + self.eps)
        if self.elementwise_affine:
            out = out * self.weight.view(1, -1, 1, 1) + self.bias.view(1, -1, 1, 1)
        return out


class TritonRMSNorm2d(nn.LayerNorm):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return TritonRMSNorm2dFunc.apply(x, self.weight, self.bias, self.eps)


class LayerNorm3d(nn.Module):
    """Channel-wise LayerNorm for 5D (B, C, D, H, W); same logic as LayerNorm2d with 5D view."""

    def __init__(self, normalized_shape: int, eps: float = 1e-5, elementwise_affine: bool = True):
        super().__init__()
        self.eps = eps
        self.elementwise_affine = elementwise_affine
        if elementwise_affine:
            self.weight = nn.Parameter(torch.ones(normalized_shape))
            self.bias = nn.Parameter(torch.zeros(normalized_shape))
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = x - torch.mean(x, dim=1, keepdim=True)
        out = out / torch.sqrt(torch.square(out).mean(dim=1, keepdim=True) + self.eps)
        if self.elementwise_affine:
            # (1, C, 1, 1, 1) for 5D
            view = (1, -1) + (1,) * (x.dim() - 2)
            out = out * self.weight.view(view) + self.bias.view(view)
        return out


class RMSNorm3d(nn.Module):
    """RMSNorm for 5D (B, C, D, H, W); channel-wise, no Triton (fallback for 3D)."""

    def __init__(self, normalized_shape: int, eps: float = 1e-6, elementwise_affine: bool = True):
        super().__init__()
        self.eps = eps
        self.elementwise_affine = elementwise_affine
        if elementwise_affine:
            self.weight = nn.Parameter(torch.ones(normalized_shape))
            self.bias = nn.Parameter(torch.zeros(normalized_shape))
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        rms = torch.sqrt(torch.mean(x * x, dim=1, keepdim=True) + self.eps)
        out = x / rms
        if self.elementwise_affine:
            view = (1, -1) + (1,) * (x.dim() - 2)
            out = out * self.weight.view(view) + self.bias.view(view)
        return out


# register normalization function here
REGISTERED_NORM_DICT: dict[str, type] = {
    "bn2d": nn.BatchNorm2d,
    "bn3d": nn.BatchNorm3d,
    "ln": nn.LayerNorm,
    "ln2d": LayerNorm2d,
    "ln3d": LayerNorm3d,
    "trms2d": TritonRMSNorm2d,
    "trms3d": RMSNorm3d,
}


def build_norm(name="bn2d", num_features=None, dims=2, **kwargs) -> Optional[nn.Module]:
    # Map 2d norm names to 3d when dims=3 (so configs can keep "trms2d" etc.)
    if dims == 3 and name in ("bn2d", "ln2d", "trms2d"):
        name = {"bn2d": "bn3d", "ln2d": "ln3d", "trms2d": "trms3d"}[name]
    if name in ["ln", "ln2d", "ln3d", "trms2d", "trms3d"]:
        kwargs["normalized_shape"] = num_features
    else:
        kwargs["num_features"] = num_features
    if name in REGISTERED_NORM_DICT:
        norm_cls = REGISTERED_NORM_DICT[name]
        args = build_kwargs_from_config(kwargs, norm_cls)
        return norm_cls(**args)
    else:
        return None


def reset_bn(
    model: nn.Module,
    data_loader: list,
    sync=True,
    progress_bar=False,
) -> None:
    import copy

    import torch.nn.functional as F
    from tqdm import tqdm

    from ..utils.network import get_device
    from ..utils.list import list_join
    
    # Simple implementations for missing utilities
    class AverageMeter:
        def __init__(self, is_distributed=False):
            self.avg = None
            self.count = 0
        
        def update(self, val, n=1):
            if self.avg is None:
                self.avg = val.clone() if isinstance(val, torch.Tensor) else val
            else:
                self.avg = (self.avg * self.count + val * n) / (self.count + n)
            self.count += n
    
    def is_master():
        return True
    
    def sync_tensor(tensor, reduce="cat"):
        return tensor

    bn_mean = {}
    bn_var = {}

    tmp_model = copy.deepcopy(model)
    for name, m in tmp_model.named_modules():
        if isinstance(m, _BatchNorm):
            bn_mean[name] = AverageMeter(is_distributed=False)
            bn_var[name] = AverageMeter(is_distributed=False)

            def new_forward(bn, mean_est, var_est):
                def lambda_forward(x):
                    x = x.contiguous()
                    if sync:
                        batch_mean = x.mean(0, keepdim=True).mean(2, keepdim=True).mean(3, keepdim=True)  # 1, C, 1, 1
                        batch_mean = sync_tensor(batch_mean, reduce="cat")
                        batch_mean = torch.mean(batch_mean, dim=0, keepdim=True)

                        batch_var = (x - batch_mean) * (x - batch_mean)
                        batch_var = batch_var.mean(0, keepdim=True).mean(2, keepdim=True).mean(3, keepdim=True)
                        batch_var = sync_tensor(batch_var, reduce="cat")
                        batch_var = torch.mean(batch_var, dim=0, keepdim=True)
                    else:
                        batch_mean = x.mean(0, keepdim=True).mean(2, keepdim=True).mean(3, keepdim=True)  # 1, C, 1, 1
                        batch_var = (x - batch_mean) * (x - batch_mean)
                        batch_var = batch_var.mean(0, keepdim=True).mean(2, keepdim=True).mean(3, keepdim=True)

                    batch_mean = torch.squeeze(batch_mean)
                    batch_var = torch.squeeze(batch_var)

                    mean_est.update(batch_mean.data, x.size(0))
                    var_est.update(batch_var.data, x.size(0))

                    # bn forward using calculated mean & var
                    _feature_dim = batch_mean.shape[0]
                    return F.batch_norm(
                        x,
                        batch_mean,
                        batch_var,
                        bn.weight[:_feature_dim],
                        bn.bias[:_feature_dim],
                        False,
                        0.0,
                        bn.eps,
                    )

                return lambda_forward

            m.forward = new_forward(m, bn_mean[name], bn_var[name])

    # skip if there is no batch normalization layers in the network
    if len(bn_mean) == 0:
        return

    tmp_model.eval()
    with torch.no_grad():
        with tqdm(total=len(data_loader), desc="reset bn", disable=not progress_bar or not is_master()) as t:
            for images in data_loader:
                images = images.to(get_device(tmp_model))
                tmp_model(images)
                t.set_postfix(
                    {
                        "bs": images.size(0),
                        "res": list_join(images.shape[-2:], "x"),
                    }
                )
                t.update()

    for name, m in model.named_modules():
        if name in bn_mean and bn_mean[name].count > 0:
            feature_dim = bn_mean[name].avg.size(0)
            assert isinstance(m, _BatchNorm)
            m.running_mean.data[:feature_dim].copy_(bn_mean[name].avg)
            m.running_var.data[:feature_dim].copy_(bn_var[name].avg)


def set_norm_eps(model: nn.Module, eps: Optional[float] = None) -> None:
    for m in model.modules():
        if isinstance(m, (nn.GroupNorm, nn.LayerNorm, _BatchNorm)):
            if eps is not None:
                m.eps = eps



import torch
import triton
import triton.language as tl

__all__ = ["TritonRMSNorm2dFunc"]


@triton.jit
def _rms_norm_2d_fwd_fused(
    X,  # pointer to the input
    Y,  # pointer to the output
    W,  # pointer to the weights
    B,  # pointer to the biases
    Rrms,  # pointer to the 1/rms
    M,
    C,
    N,
    num_blocks,  # number of columns in X
    eps,  # epsilon to avoid division by zero
    BLOCK_SIZE: tl.constexpr,
):
    # Map the program id to the row of X and Y it should compute.
    m_n = tl.program_id(0)
    m, n = m_n // num_blocks, m_n % num_blocks

    Y += m * C * N
    X += m * C * N
    # Compute mean

    cols = n * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = cols < N

    x_sum_square = tl.zeros([BLOCK_SIZE], dtype=tl.float32)
    for off in range(0, C):
        x = tl.load(X + off * N + cols, mask=mask, other=0.0).to(tl.float32)
        x_sum_square += x * x
    mean_square = x_sum_square / C
    rrms = 1 / tl.sqrt(mean_square + eps)
    # Write rstd
    tl.store(Rrms + m * N + cols, rrms, mask=mask)
    # Normalize and apply linear transformation
    for off in range(0, C):
        pos = off * N + cols
        w = tl.load(W + off)
        b = tl.load(B + off)
        x = tl.load(X + pos, mask=mask, other=0.0).to(tl.float32)
        x_hat = x * rrms
        y = x_hat * w + b
        # Write output
        tl.store(Y + pos, y, mask=mask)


@triton.jit
def _rms_norm_2d_bwd_dx_fused(
    DX,  # pointer to the input gradient
    DY,  # pointer to the output gradient
    DW,  # pointer to the partial sum of weights gradient
    DB,  # pointer to the partial sum of biases gradient
    X,  # pointer to the input
    W,  # pointer to the weights
    B,  # pointer to the biases
    Rrms,  # pointer to the 1/rms
    M,
    C,
    N,  # number of columns in X
    num_blocks,
    eps,  # epsilon to avoid division by zero
    GROUP_SIZE_M: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
    BLOCK_SIZE_C: tl.constexpr,
):
    # Map the program id to the elements of X, DX, and DY it should compute.
    m_n = tl.program_id(0)
    m, n = m_n // num_blocks, m_n % num_blocks
    X += m * C * N
    DY += m * C * N
    DX += m * C * N
    Rrms += m * N

    cols = n * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = cols < N
    # Offset locks and weights/biases gradient pointer for parallel reduction
    DW = DW + m_n * C
    DB = DB + m_n * C
    rrms = tl.load(Rrms + cols, mask=mask, other=1)
    # Load data to SRAM
    c1 = tl.zeros([BLOCK_SIZE], dtype=tl.float32)
    for off in range(0, C):
        pos = off * N + cols
        x = tl.load(X + pos, mask=mask, other=0).to(tl.float32)
        dy = tl.load(DY + pos, mask=mask, other=0).to(tl.float32)
        w = tl.load(W + off).to(tl.float32)
        # Compute dx
        xhat = x * rrms
        wdy = w * dy
        xhat = tl.where(mask, xhat, 0.0)
        wdy = tl.where(mask, wdy, 0.0)
        c1 += xhat * wdy
        # Accumulate partial sums for dw/db
        tl.store(DW + off, tl.sum((dy * xhat).to(w.dtype), axis=0))
        tl.store(DB + off, tl.sum(dy.to(w.dtype), axis=0))

    c1 /= C
    for off in range(0, C):
        pos = off * N + cols
        x = tl.load(X + pos, mask=mask, other=0).to(tl.float32)
        dy = tl.load(DY + pos, mask=mask, other=0).to(tl.float32)
        w = tl.load(W + off).to(tl.float32)
        xhat = x * rrms
        wdy = w * dy
        dx = (wdy - (xhat * c1)) * rrms
        # Write dx
        tl.store(DX + pos, dx, mask=mask)


class TritonRMSNorm2dFunc(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, weight, bias, eps):
        # allocate output
        y = torch.empty_like(x)
        # reshape input data into 2D tensor
        x_arg = x.reshape(x.shape[0], x.shape[1], -1)
        M, C, N = x_arg.shape
        rrms = torch.empty((M, N), dtype=torch.float32, device="cuda")
        # Less than 64KB per feature: enqueue fused kernel
        BLOCK_SIZE = 256
        num_blocks = triton.cdiv(N, BLOCK_SIZE)
        num_warps = 8
        # enqueue kernel
        _rms_norm_2d_fwd_fused[(M * num_blocks,)](  #
            x_arg,
            y,
            weight,
            bias,
            rrms,  #
            M,
            C,
            N,
            num_blocks,
            eps,  #
            BLOCK_SIZE=BLOCK_SIZE,
            num_warps=num_warps,
            num_ctas=1,
        )
        ctx.save_for_backward(x, weight, bias, rrms)
        ctx.BLOCK_SIZE = BLOCK_SIZE
        ctx.num_blocks = num_blocks
        ctx.num_warps = num_warps
        ctx.eps = eps
        return y

    @staticmethod
    def backward(ctx, dy):
        x, w, b, rrms = ctx.saved_tensors
        num_blocks = ctx.num_blocks

        x_arg = x.reshape(x.shape[0], x.shape[1], -1)
        M, C, N = x_arg.shape
        # GROUP_SIZE_M = 64
        GROUP_SIZE_M = M * num_blocks
        # allocate output
        _dw = torch.empty((GROUP_SIZE_M, C), dtype=x.dtype, device=w.device)
        _db = torch.empty((GROUP_SIZE_M, C), dtype=x.dtype, device=w.device)
        dw = torch.empty((C,), dtype=w.dtype, device=w.device)
        db = torch.empty((C,), dtype=w.dtype, device=w.device)
        dx = torch.empty_like(dy)
        # enqueue kernel using forward pass heuristics
        # also compute partial sums for DW and DB
        # print(f"M={M}, num_blocks={num_blocks}, dx={dx.shape}, dy={dy.shape}, _dw={_dw.shape}, _db={_db.shape}, x={x.shape}, w={w.shape}, b={b.shape}, m={m.shape}, v={v.shape}, M={M}, C={C}, N={N}")
        _rms_norm_2d_bwd_dx_fused[(M * num_blocks,)](  #
            dx,
            dy,
            _dw,
            _db,
            x,
            w,
            b,
            rrms,  #
            M,
            C,
            N,
            num_blocks,
            ctx.eps,  #
            BLOCK_SIZE=ctx.BLOCK_SIZE,
            GROUP_SIZE_M=GROUP_SIZE_M,  #
            BLOCK_SIZE_C=triton.next_power_of_2(C),
            num_warps=ctx.num_warps,
        )
        dw = _dw.sum(dim=0)
        db = _db.sum(dim=0)
        return dx, dw, db, None



# BERT architecture for the Masked Bidirectional Encoder Transformer
import torch
from torch import nn


def param_count(archi, model):
    print(f"Size of model {archi}: "
          f"{sum(p.numel() for p in model.parameters() if p.requires_grad) / 10 ** 6:.3f}M")


class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim, dropout=0.):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, hidden_dim),
            nn.GELU(approximate="tanh"),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, dim),
            nn.Dropout(dropout)
        )

    def forward(self, x):
        return self.net(x)


class Attention(nn.Module):
    def __init__(self, embed_dim, num_heads, dropout=0.):
        super(Attention, self).__init__()
        self.dim = embed_dim
        self.mha = nn.MultiheadAttention(embed_dim, num_heads=num_heads, dropout=dropout, batch_first=True, bias=True)

    def forward(self, x, cond=None):
        if cond is not None:
            x = self.mha(x, cond, cond, need_weights=False)[0]
        else:
            x = self.mha(x, x, x, need_weights=False)[0]
        return x


class NormLayer(nn.Module):
    def __init__(self, x_dim):
        super().__init__()
        self.norm_final = nn.LayerNorm(x_dim, eps=1e-6)

    def forward(self, x):
        x = self.norm_final(x)
        return x


class Block(nn.Module):
    def __init__(self, hidden_dim, cond_dim, heads, mlp_dim, dropout=0.):
        super().__init__()
        self.mlp = nn.Sequential(nn.Linear(cond_dim, hidden_dim),
                                 nn.LayerNorm(hidden_dim))

        self.ln1 = nn.LayerNorm(hidden_dim, eps=1e-6)
        self.self_attn = Attention(hidden_dim, heads, dropout=dropout)

        self.ln2 = nn.LayerNorm(hidden_dim, eps=1e-6)
        self.cross_attn = Attention(hidden_dim, heads, dropout=dropout)

        self.ln3 = nn.LayerNorm(hidden_dim, eps=1e-6)
        self.ff = FeedForward(hidden_dim, mlp_dim, dropout=dropout)

    def forward(self, x, cond):
        cond = self.mlp(cond)
        x = x + self.self_attn(self.ln1(x))
        x = x + self.cross_attn(self.ln2(x), cond)
        x = x + self.ff(self.ln3(x))
        return x


class TransformerEncoder(nn.Module):
    def __init__(self, hidden_dim, cond_dim, depth, heads, mlp_dim, dropout=0.):
        super().__init__()
        self.layers = nn.ModuleList([])
        for layer in range(depth):
            self.layers.append(
                Block(hidden_dim, cond_dim, heads, mlp_dim, dropout=dropout)
            )

    def forward(self, x, cond):
        for block in self.layers:
            x = block(x, cond)
        return x


class MaskGIT_TXT(nn.Module):
    def __init__(self, input_size=16, c=768, hidden_dim=768, cond_dim=768, codebook_size=1024,
                 depth=12, heads=16, mlp_dim=3072, dropout=0., proj=1):
        super().__init__()

        self.c = c
        self.proj = proj
        self.input_size = input_size
        self.hidden_dim = hidden_dim
        self.codebook_size = codebook_size

        self.tok_emb = nn.Embedding(codebook_size+1, c)
        self.pos_emb = nn.Embedding(self.input_size ** 2, c)

        self.first_norm = nn.LayerNorm(c, eps=1e-6)
        self.in_proj = nn.Conv2d(c, hidden_dim, kernel_size=proj, stride=proj, bias=False)

        self.transformer = TransformerEncoder(hidden_dim=hidden_dim, cond_dim=cond_dim, depth=depth, heads=heads, mlp_dim=mlp_dim, dropout=dropout)

        self.out_proj = nn.Sequential(nn.Conv2d(hidden_dim, c*(proj**2), kernel_size=1, stride=1, bias=False),
                                      nn.PixelShuffle(proj)
                                      )

        self.last_norm = nn.LayerNorm(c, eps=1e-6)
        self.bias = nn.Parameter(torch.zeros(1, 1, codebook_size+1))

        self.initialize_weights()

    def initialize_weights(self):
        # Initialize transformer layers:
        def _basic_init(module):
            if isinstance(module, nn.Linear):
                nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    nn.init.constant_(module.bias, 0)

        self.apply(_basic_init)

        # Init embedding
        nn.init.normal_(self.tok_emb.weight, mean=0.0, std=0.02)
        nn.init.normal_(self.pos_emb.weight, mean=0.0, std=0.02)

        # Init projection in
        nn.init.xavier_uniform_(self.in_proj.weight)

        # Zero-out cross attn projection layers in blocks:
        for block in self.transformer.layers:
            nn.init.constant_(block.mlp[0].weight, 0)
            nn.init.constant_(block.mlp[0].bias, 0)

        # Init projection out
        nn.init.xavier_uniform_(self.out_proj[0].weight)

    def partially_init_from_pretrained(self, ckpt):
        pretrained_model = ckpt['model_state_dict']
        print("Copy only transformer weights from pretrained model")
        for source_parameter, target_parameter in zip(pretrained_model.keys(), self.state_dict().keys()):
            if source_parameter == target_parameter and \
                    self.state_dict()[target_parameter].size() == pretrained_model[source_parameter].size()\
                    and "transformer" in source_parameter:
                print("copying:", source_parameter, self.state_dict()[target_parameter].size())
                self.state_dict()[target_parameter].data.copy_(pretrained_model[source_parameter])

    def forward(self, x, y, drop_label=None):
        b, w, h = x.size()

        # Drop the text-label y --> b, seq_len, cond_dim
        if drop_label is not None:
            y = torch.where(drop_label.view(-1, 1, 1), torch.zeros_like(y), y)

        # position embedding
        pos = torch.arange(0, w*h, dtype=torch.long, device=x.device)  # shape (t)
        pos = self.pos_emb(pos)

        x = x.view(b, -1)
        x = self.tok_emb(x) + pos                                            # b, w*h, c

        x = self.first_norm(x)

        # reshape, proj to smaller space, reshape (patchify!)
        x = x.transpose(1, 2).contiguous().view(b, self.c, w, h)                         # b, c, w, h
        x = self.in_proj(x)                                                              # b, hidden, w // proj, h // proj
        x = x.view(b, self.hidden_dim, -1).transpose(1, 2).contiguous()                  # b, (w // proj * h // proj), hidden

        x = self.transformer(x, y)                                                       # b, (w // proj * h // proj), hidden

        x = x.transpose(1, 2).contiguous().view(b, self.hidden_dim, w//self.proj, h//self.proj)  # b, hidden, w // proj, h // proj
        x = self.out_proj(x)                                                               # b, hidden//proj**2, w, h
        x = x.view(b, self.c, -1).transpose(1, 2).contiguous()                             # b, w * h, hidden//proj**2

        x = self.last_norm(x) # normalize before final prediction

        logit = torch.matmul(x, self.tok_emb.weight.T) + self.bias

        return logit

